# -*- coding: utf-8 -*-


import io
import boto3

from dp_tornado.engine.helper import Helper as dpHelper


class S3Helper(dpHelper):
    def upload(self,
               access_key_id,
               secret_access_key,
               region_name,
               bucket_name,
               src,
               dest,
               uploaded_check=True,
               ExtraArgs=None):
        if self.helper.misc.type.check.string(src):
            if not self.helper.io.path.is_file(src):
                return False
        elif self.helper.misc.system.py_version <= 2 and not isinstance(src, file):
            return False
        elif self.helper.misc.system.py_version >= 3 and not isinstance(src, io.IOBase):
            return False

        s3 = boto3.client(
            service_name='s3',
            region_name=region_name,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key)

        if self.helper.misc.type.check.string(src):
            s3.upload_file(src, bucket_name, dest, ExtraArgs=ExtraArgs)
        elif src:
            s3.upload_fileobj(src, bucket_name, dest, ExtraArgs=ExtraArgs)

        if uploaded_check:
            return self.exists(
                access_key_id=access_key_id,
                secret_access_key=secret_access_key,
                region_name=region_name,
                bucket_name=bucket_name,
                key=dest)

        return True

    def download(self,
                 access_key_id,
                 secret_access_key,
                 region_name,
                 bucket_name,
                 src,
                 dest,
                 **kwargs):
        s3 = boto3.client(
            service_name='s3',
            region_name=region_name,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key)

        self.helper.io.path.mkdir(self.helper.io.path.dirname(dest))
        s3.download_file(bucket_name, src, dest, ExtraArgs=kwargs)

        return True if self.helper.io.path.is_file(dest) else False

    def copy(self,
             access_key_id,
             secret_access_key,
             region_name,
             src_bucket_name,
             src_key,
             dest_bucket_name,
             dest_key,
             copied_check=True):
        s3 = boto3.client(
            service_name='s3',
            region_name=region_name,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key)

        copy_source = {
            'Bucket': src_bucket_name,
            'Key': src_key
        }

        s3.copy(copy_source, dest_bucket_name, dest_key)

        if copied_check:
            return self.exists(
                access_key_id=access_key_id,
                secret_access_key=secret_access_key,
                region_name=region_name,
                bucket_name=dest_bucket_name,
                key=dest_key)

        return True

    def browse(self, access_key_id, secret_access_key, region_name, bucket_name, prefix):
        try:
            s3 = boto3.client(
                service_name='s3',
                region_name=region_name,
                aws_access_key_id=access_key_id,
                aws_secret_access_key=secret_access_key)

            result = s3.list_objects(Bucket=bucket_name, Prefix=prefix)

            if 'Contents' not in result:
                return []

            contents = []

            for e in result['Contents']:
                if 'Key' not in e or 'Size' not in e:
                    continue

                attrs = {
                    'Size': e['Size'],
                    'LastModified': e['LastModified'] if 'LastModified' in e else None,
                    'StorageClass': e['StorageClass'] if 'StorageClass' in e else None,
                    'ETag': e['ETag'] if 'ETag' in e else None
                }

                contents.append((e['Key'], attrs))

            return contents

        except Exception as e:
            self.logging.exception(e)

            return False

    def exists(self, key, access_key_id, secret_access_key, region_name, bucket_name):
        s3 = boto3.client(
            service_name='s3',
            region_name=region_name,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key)

        try:
            return True if s3.head_object(Bucket=bucket_name, Key=key) else False
        except Exception:
            return False

    def remove(self, access_key_id, secret_access_key, region_name, bucket_name, key=None, prefix=None, deleted_check=True):
        assert key or prefix
        assert not (key and prefix)

        try:
            s3 = boto3.client(
                service_name='s3',
                region_name=region_name,
                aws_access_key_id=access_key_id,
                aws_secret_access_key=secret_access_key)

            # with specified key, single key
            if key:
                deleted = s3.delete_object(Bucket=bucket_name, Key=key)

                # Check response
                if not deleted or 'ResponseMetadata' not in deleted \
                        or 'HTTPStatusCode' not in deleted['ResponseMetadata'] \
                        or deleted['ResponseMetadata']['HTTPStatusCode'] % 200 > 99:
                    return False

                if deleted_check:
                    if self.exists(
                            access_key_id=access_key_id,
                            secret_access_key=secret_access_key,
                            region_name=region_name,
                            bucket_name=bucket_name,
                            key=key):
                        return False

                return key,

            # with prefix, multiple keys
            elif prefix:
                keys = self.browse(
                    access_key_id=self.ini.static.aws_id,
                    secret_access_key=self.ini.static.aws_secret,
                    region_name=self.ini.static.aws_region,
                    bucket_name=self.ini.static.aws_bucket,
                    prefix=prefix)

                if keys is False:
                    return False

                elif not keys:
                    return None

                delete = {'Objects': [{'Key': e[0]} for e in keys]}
                deleted = s3.delete_objects(Bucket=bucket_name, Delete=delete)

                # Check response
                if not deleted or 'Deleted' not in deleted \
                        or 'ResponseMetadata' not in deleted \
                        or 'HTTPStatusCode' not in deleted['ResponseMetadata'] \
                        or deleted['ResponseMetadata']['HTTPStatusCode'] % 200 > 99:
                    return False

                return tuple([e['Key'] for e in deleted['Deleted'] if 'Key' in e])

            return False

        except Exception as e:
            self.logging.exception(e)

            return False

    def generate_presigned_post(self,
                                access_key_id,
                                secret_access_key,
                                region_name,
                                bucket_name,
                                key,
                                success_action_redirect=None,
                                content_length_range=None,
                                max_content_length=None,
                                expires_in=6000,
                                acl=None):
        s3 = boto3.client(
            service_name='s3',
            region_name=region_name,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key)

        fields = {}
        conditions = []

        if success_action_redirect:
            fields['success_action_redirect'] = success_action_redirect
            conditions.append({'success_action_redirect': success_action_redirect})
        else:
            fields['success_action_status'] = '201'
            conditions.append({'success_action_status': '201'})

        if acl:
            conditions.append({'acl': acl})

        if content_length_range and len(content_length_range) == 2:
            conditions.append(["content-length-range", content_length_range[0], max_content_length])
        elif max_content_length:
            conditions.append(["content-length-range", 0, max_content_length])

        payload = s3.generate_presigned_post(
            Bucket=bucket_name,
            Key=key,
            Fields=fields,
            Conditions=conditions,
            ExpiresIn=expires_in)

        return {
            'action': payload['url'],
            'fields': [{'name': k, 'value': v} for k, v in payload['fields'].items()]
        }



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class FormHelper(dpHelper):
    def validate(self, controller, fields, error_res='json'):
        assert error_res in ('json', 'http', 'code')

        output = {}

        missing_reason = 'missing'
        invalid_reason = 'invalid'

        for field, payload in fields.items():
            missing_message = payload['missing'] if 'missing' in payload else None
            invalid_message = payload['invalid'] if 'invalid' in payload else None
            cast = payload['cast'] if 'cast' in payload else None

            value = controller.get_argument(name=field, **payload)

            # Value validation
            if (cast is bool and value == -1) or (cast is not bool and value is False):
                return self._validate_response(controller, error_res, field, invalid_reason, invalid_message)

            # Check required
            if value is None and 'required' in payload and payload['required']:
                return self._validate_response(controller, error_res, field, missing_reason, missing_message)

            output[field] = value

        return output

    def _validate_response(self, controller, error_res, field, reason, message):
        if error_res == 'json':
            return controller.finish({
                'result': False,
                'error': {
                    'field': field,
                    'reason': reason,
                    'message': message
                }
            })

        return controller.finish_with_error(400, message)



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import time as abs_time


class TimestampHelper(dpHelper):
    def now(self, ms=False):
        return self.helper.numeric.cast.long(abs_time.time() if not ms else round(abs_time.time() * 1000))

    def yesterday(
            self,
            auto=None,
            timestamp=None,
            datetime=None,
            timezone=None,
            yyyymmdd=None,
            yyyymmddhhiiss=None,
            ms=False):
        timestamp = self.convert(
            auto=auto,
            timestamp=timestamp,
            datetime=datetime,
            timezone=timezone,
            yyyymmdd=yyyymmdd,
            yyyymmddhhiiss=yyyymmddhhiiss,
            ms=ms)

        return timestamp - (3600*24*(1 if not ms else 1000))

    def tommorow(
            self,
            auto=None,
            timestamp=None,
            datetime=None,
            timezone=None,
            yyyymmdd=None,
            yyyymmddhhiiss=None,
            ms=False):
        timestamp = self.convert(
            auto=auto,
            timestamp=timestamp,
            datetime=datetime,
            timezone=timezone,
            yyyymmdd=yyyymmdd,
            yyyymmddhhiiss=yyyymmddhhiiss,
            ms=ms)

        return timestamp + (3600*24*(1 if not ms else 1000))

    def to_datetime(self, *args, **kwargs):
        return self.helper.datetime.from_timestamp(*args, **kwargs)

    def mktime(self, year=1970, month=1, day=1, hour=0, minute=0, second=0, microsecond=0, ms=False):
        p_tuple = (
            year,
            month,
            day,
            hour,
            minute,
            second,
            0,
            0,
            0
        )

        timestamp = self.helper.numeric.cast.long(abs_time.mktime(p_tuple))

        if not ms:
            return timestamp
        else:
            return (timestamp * 1000) + microsecond

    def from_datetime(self, datetime, ms=False):
        return self.mktime(
            year=datetime.year,
            month=datetime.month,
            day=datetime.day,
            hour=datetime.hour,
            minute=datetime.minute,
            second=datetime.second,
            microsecond=datetime.microsecond // 1000,
            ms=ms)

    def convert(
            self,
            auto=None,
            datetime=None,
            timezone=None,
            timestamp=None,
            yyyymmdd=None,
            yyyymmddhhiiss=None,
            ms=False):
        datetime = self.helper.datetime.convert(
            auto=auto,
            datetime=datetime,
            timezone=timezone,
            timestamp=timestamp,
            yyyymmdd=yyyymmdd,
            yyyymmddhhiiss=yyyymmddhhiiss,
            ms=ms)

        return self.from_datetime(datetime=datetime, ms=ms)



# -*- coding: utf-8 -*-


from dp_tornado.engine.controller import Controller as dpController


class ListController(dpController):
    def get(self, article_id=None):
        if article_id:
            return self.get_view(article_id)

        validate = self.helper.validator.form.validate(self, {
            'page': {'default': 1, 'cast': self.helper.misc.type.int, 'required': False},
            'rpp': {'default': 10, 'cast': self.helper.misc.type.int, 'required': False}
        }, error_res='code')

        if not validate:
            return False

        params = {
            'articles': self.model.article.inquiry(**validate),
        }

        self.render('bbs/list.html', params)

    def get_view(self, article_id):
        article = self.model.article.inquiry(article_id=article_id, rpp=1)

        if not article:
            return self.finish_with_error(404)

        params = {
            'request_uri': self.get_argument('request_uri'),
            'article': article
        }

        self.render('bbs/view.html', params)



# -*- coding: utf-8 -*-


import os
import sys
import logging

from dp_tornado.engine.static_handler import StaticHandler
from dp_tornado.engine.engine import EngineSingleton as dpEngineSingleton
from dp_tornado.version import __version__


engine = dpEngineSingleton()
logging.basicConfig(format='[%(asctime)s][%(levelname)s] %(message)s')


class Bootstrap(object):
    @staticmethod
    def init_template(engine_path, application_path, template_name):
        valid = True

        for e in ('config', 'controller', 'model', 'helper'):
            if os.path.isdir(os.path.join(application_path, e)):
                valid = False
                break

        if not valid:
            logging.warning('Default directory structure initialzation skipped. Directory not empty.')
            return False

        else:
            import shutil

            template_path = os.path.join(engine_path, 'engine', 'template', template_name)

            if not os.path.isdir(template_path):
                return False

            for root, dirs, files in os.walk(template_path):
                path = root[len(template_path)+1:]
                app_path = os.path.join(application_path, path)

                if path and os.path.isdir(app_path):
                    continue

                if not os.path.isdir(app_path):
                    os.mkdir(app_path)

                for file in files:
                    src = os.path.join(root, file)
                    dest = os.path.join(app_path, file)

                    if not os.path.isfile(dest):
                        if src.endswith('.pyc'):
                            continue

                        shutil.copy(src, dest)

            # default requirements.txt file.
            with open(os.path.join(application_path, 'requirements.txt'), 'w') as fp:
                fp.write('dp-tornado==%s' % __version__)

            return True

    @staticmethod
    def init_args():
        import argparse

        parser = argparse.ArgumentParser()

        parser.add_argument('--app-path', help='App Path')
        parser.add_argument('--mode', help='Mode')

        parser.add_argument('-i', '--identifier', help='Identifier')
        parser.add_argument('-p', '--port', type=int, help='Binding port')

        return parser.parse_known_args()

    @staticmethod
    def init_ini(application_path, ini_file, as_cli=False):
        args, args_unkonwn = Bootstrap.init_args()

        # App Options

        if args.mode:
            engine.ini.app.set('mode', args.mode)

        engine.ini.app.get('mode', default='NOT SET')

        # Specify Sandbox Mode
        if (engine.ini.app.mode or '').lower() in ('debug', 'sandbox', 'test', 'dev', 'development', 'develop'):
            pass

        # Specify Production Mode
        elif (engine.ini.app.mode or '').lower() in ('prod', 'production', 'real', 'service'):
            pass

        combined_path = engine.ini.static.get('path', default='combined')
        static_prefix = engine.ini.static.get('prefix', default='/s/')
        static_minify = engine.ini.static.get('minify', default=True)

        if combined_path.find('{server_name}') != -1:
            import socket
            combined_path = combined_path.replace('{server_name}', socket.gethostname())

        if static_prefix[-1] != '/':
            static_prefix = '%s/' % static_prefix

        if combined_path[0] == '/':
            combined_path = combined_path[1:]

        if combined_path[-1] == '/':
            combined_path = combined_path[:-1]

        combined_prefix = '%s%s/' % (static_prefix, combined_path)
        combined_url = combined_path
        combined_path = os.path.join(application_path, 'static', combined_path)

        engine.ini.static.set('combined_prefix', combined_prefix)
        engine.ini.static.set('combined_url', combined_url)
        engine.ini.static.set('combined_path', combined_path)
        engine.ini.static.set('path_sys', os.path.join(application_path, 'static'))

        engine.ini.server.set('application_path', application_path)
        engine.ini.server.set('python', sys.executable)

        if args.port:
            engine.ini.server.set('port', args.port)

        # Identifier
        if args.identifier:
            engine.ini.server.set('identifier', args.identifier)
        else:
            engine.ini.server.set('identifier', engine.helper.misc.uuid.v1())

        # Setup Options
        engine.ini.server.get('max_worker', default=1)
        engine.ini.server.get('num_processes', default=0)
        engine.ini.server.get('port', default=8080)

        # Enabled production mode when running via cli.
        if not as_cli:
            engine.ini.server.get('debug', default=False)
        else:
            engine.ini.server.set('debug', False)

        engine.ini.server.get('gzip', default=True)
        engine.ini.crypto.get('key', default='CR$t0-$CR@T')
        engine.ini.session.get('dsn', default=None)
        engine.ini.session.get('expire_in', default=7200)
        engine.ini.server.get('max_body_size', default=1024*1024*10)

        # If enabled debugging mode, disable process forking.
        if engine.ini.server.debug:
            engine.ini.server.set('num_processes', 1)

        m17n = engine.ini.server.get('m17n', default='').strip()
        m17n = m17n.split(',') if m17n else ['dummy']
        m17n = [e.strip() for e in m17n]

        engine.ini.server.set('m17n', m17n)

        # Static AWS
        engine.ini.static.get('aws_id')
        engine.ini.static.get('aws_secret')
        engine.ini.static.get('aws_bucket')
        engine.ini.static.get('aws_region')
        engine.ini.static.get('aws_endpoint')

        # Scheduler
        engine.ini.scheduler.get('timezone', default='')
        engine.ini.scheduler.get('mode', default='web')

        exception_delegate = engine.ini.logging.get('exception_delegate', default='') or None
        access_logging = engine.ini.logging.get('access', default=False)
        sql_logging = engine.ini.logging.get('sql', default=False)
        aws_logging = engine.ini.logging.get('aws', default=False)
        http_logging = engine.ini.logging.get('http', default=False)
        dp_logging = engine.ini.logging.get('dp', default=30 if not engine.ini.server.debug else 10)

        if exception_delegate:
            try:
                exception_delegate = eval('engine.%s' % exception_delegate)
                engine.logger.set_delegate_handler(exception_delegate)
            except AttributeError:
                logging.error('The specified exception delegate is invalid.')
                exception_delegate = None

        engine.ini.logging.set('exception_delegate', exception_delegate)

        # Access Logging
        import tornado.log
        engine.logger.set_level(tornado.log.access_log.name, logging.DEBUG if access_logging else logging.WARN)

        # dp Logging
        engine.logger.set_level(engine.logger.default_logger_name, dp_logging or 100)

        # SQLAlchemy logging level
        if sql_logging:
            logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)

        # AWS logging level (boto
        if not aws_logging:
            logging.getLogger('boto3').setLevel(logging.WARNING)
            logging.getLogger('botocore').setLevel(logging.WARNING)
            logging.getLogger('nose').setLevel(logging.WARNING)
            logging.getLogger('s3transfer').setLevel(logging.WARNING)

        # Requests logging level
        if not http_logging:
            logging.getLogger('requests').setLevel(logging.WARNING)

        # Disable Pillow log
        logging.getLogger('PIL.PngImagePlugin').setLevel(logging.WARNING)

        return {
            'template_path': os.path.join(application_path, 'view'),
            'static_path': os.path.join(application_path, 'static'),
            'static_handler_class': StaticHandler,
            'static_url_prefix': static_prefix,
            'static_minify': static_minify,
            'static_combined_url': combined_url,
            'combined_static_path': combined_path,
            'combined_static_url_prefix': combined_prefix,
            'debug': engine.ini.server.debug,
            'gzip': engine.ini.server.gzip,
            'cookie_secret': engine.ini.server.get('cookie_secret', default='default_cookie_secret')
        }



# -*- coding: utf-8 -*-


import httpagentparser

from collections import namedtuple


Device = namedtuple('Device', ['name', 'version', 'major', 'css', 'raw'])
OS = namedtuple('OS', ['name', 'version', 'major', 'css', 'raw'])
Browser = namedtuple('Browser', ['name', 'version', 'major', 'css', 'raw'])


class UserAgent(object):
    def __init__(self, ua):
        ua = httpagentparser.detect(ua)

        if 'dist' in ua:
            ua['d'] = ua['dist']
            del ua['dist']

        if 'platform' in ua:
            ua['p'] = ua['platform']
            del ua['platform']

        if 'os' in ua:
            ua['o'] = ua['os']
            del ua['os']

        if 'browser' in ua:
            ua['b'] = ua['browser']
            del ua['browser']

        self._ua = ua
        self.is_bot = True if 'bot' in ua and ua['bot'] else False

    @property
    def device(self):
        _device = getattr(self, '_device', None)

        if _device:
            return _device

        raw = self._ua['d'] if 'd' in self._ua else None
        family = self._ua['d']['name'] if 'd' in self._ua and 'name' in self._ua['d'] else None
        version = self._ua['d']['version'] if 'd' in self._ua and 'version' in self._ua['d'] else None

        try:
            major = int(float(version.split('.')[0])) if version else 0
        except ValueError:
            major = 0

        family = family or 'Unknown'
        version = version or 'Unknown'

        css = [('_device-%s-%s' % (family, version)).lower().replace(' ', '_')]

        if major:
            css.append(('_device-%s-%s' % (family, major)).lower().replace(' ', '_'))

        self._device = Device(family, version, major, css, raw)
        return self._device

    @property
    def os(self):
        _os = getattr(self, '_os', None)

        if _os:
            return _os

        raw = None
        family = None
        version = None

        for k in ('p', 'o'):
            raw = self._ua[k] if k in self._ua else None
            family = self._ua[k]['name'] if k in self._ua and 'name' in self._ua[k] else None
            version = self._ua[k]['version'] if k in self._ua and 'version' in self._ua[k] else None

            if k in self._ua:
                break

        if family == 'Mac OS' and version.startswith('X '):
            family = 'Mac OS X'
            version = version[2:]
        elif family == 'Windows' and version.startswith('NT '):
            family = 'Windows NT'
            version = version[3:]

        try:
            major = int(float(version.split('.')[0])) if version else 0
        except ValueError:
            major = 0

        family = family or 'Unknown'
        version = version or 'Unknown'

        css = [('_os-%s-%s' % (family, version)).lower().replace(' ', '_')]

        if major:
            css.append(('_os-%s-%s' % (family, major)).lower().replace(' ', '_'))

        self._os = OS(family, version, major, css, raw)
        return self._os

    @property
    def browser(self):
        _browser = getattr(self, '_browser', None)

        if _browser:
            return _browser

        raw = self._ua['b'] if 'b' in self._ua else None
        family = self._ua['b']['name'] if 'b' in self._ua and 'name' in self._ua['b'] else None
        version = self._ua['b']['version'] if 'b' in self._ua and 'version' in self._ua['b'] else None

        try:
            major = int(float(version.split('.')[0])) if version else 0
        except ValueError:
            major = 0

        family = family or 'Unknown'
        version = version or 'Unknown'

        css = [('_browser-%s-%s' % (family, version)).lower().replace(' ', '_').replace('.', '-')]

        if major:
            css.append(('_browser-%s-%s' % (family, major)).lower().replace(' ', '_'))

        self._browser = Browser(family, version, major, css, raw)
        return self._browser

    @property
    def css(self):
        return ' '.join(self.os.css + self.browser.css + self.device.css)

    def __str__(self):
        s = []

        if self.is_bot:
            s.append('Bot')

        s.append('%s. %s (%s)' % (self.os.name, self.os.version, self.os.major))
        s.append('%s. %s (%s)' % (self.browser.name, self.browser.version, self.browser.major))
        s.append('%s. %s (%s)' % (self.device.name, self.device.version, self.device.major))

        return ' / '.join(s)



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class EmailHelper(dpHelper):
    def validate(self, e):
        return self.helper.validator.email.validate(e)

    def username_from_email(self, email):
        pos = email.find('@')

        if pos != -1:
            return email[:pos]
        else:
            return None

    def send(self, *args, **kwargs):
        return self.helper.web.email.smtp.send(*args, **kwargs)



# -*- coding: utf-8 -*-
"""The model directly manages the data, logic, and rules of the application. `<Wikipedia>
<https://en.wikipedia.org/wiki/Model–view–controller>`_

Here is a `foo_bar` model example:

.. testcode::

    from dp_tornado.engine.model import Model as dpModel

    class FooBarModel(dpModel):
        def func1(self):
            \"""
            assert self.helper.foo.func1(10, 20) == None
            \"""
            return None

        def func2(self, a):
            \"""
            assert self.helper.foo.func2(10) == 10
            \"""
            return a

        def func3(self, a, b):
            \"""
            assert self.helper.foo.func3(10, 20) == 30
            \"""
            return a + b


File/Class Invoke rules
-----------------------
* */model/__init__.py*, **DO NOT IMPLEMENT ANY CODE IN THIS FILE**
* */model/blog/__init__.py*, ``BlogModel`` > **model.blog**
* */model/blog/admin/__init__.py*, ``AdminModel`` > **model.blog.admin**
* */model/blog/post.py*, ``PostModel`` > **model.blog.post**
* */model/blog/view.py*, ``ViewModel`` > **model.blog.view**
* */model/foo_bar.py*, ``FooBarModel`` > **model.foo_bar**


Method Invoke rules
-------------------
* */model/foo.py*, ``def func1(self)``: **model.foo.func1()**
* */model/foo.py*, ``def func2(self, a)``: **model.foo.func2(a)**
* */model/foo.py*, ``def func3(self, a, b)``: **model.foo.func3(a, b)**
"""

from .engine import Engine as dpEngine
from .singleton import Singleton as dpSingleton
from .loader import Loader as dpLoader

from sqlalchemy import create_engine
from sqlalchemy.pool import QueuePool

import threading


class InValueModelConfig(object):
    def __init__(self, driver=None, database=None, host=None, user=None, password=None, port=None, path=None, pure=None):
        self.driver = driver
        self.database = database
        self.host = host
        self.user = user
        self.password = password
        self.port = port
        self.path = path
        self.pure = pure

    def __getattr__(self, name):
        try:
            attr = self.__getattribute__(name)
        except AttributeError:
            attr = None

        return attr

    def __str__(self):
        return 'InValueModelConfig (%s://%s:%s@%s:%s/%s/%s) (Pure:%s)' \
               % (self.driver, self.user, self.password, self.host, self.port, self.database, self.path, self.pure)


class ModelSingleton(dpEngine, dpSingleton):
    _lock = threading.Lock()

    @property
    def engines(self):
        if not hasattr(self, '_engines'):
            self._engines = {}

        return self._engines

    def _parse_config(self, config_dsn, delegate, cache=False):
        if isinstance(config_dsn, InValueModelConfig):
            delegate['config'] = 'InValueModelConfig'
            delegate['database'] = config_dsn.database
            delegate['key'] = '%s_%s' % (config_dsn.driver, config_dsn.database)

            return config_dsn

        config_dsn = config_dsn.split('/')

        config = config_dsn[0]
        database = config_dsn[1]

        delegate['config'] = config
        delegate['database'] = database
        delegate['key'] = '%s_%s' % (delegate['config'], delegate['database'])

        if delegate['key'] in ModelSingleton().engines:
            return True

        try:
            package = config.split('.')
            conf = self.config

            for p in package:
                conf = conf.__getattr__(p)

            if not cache:
                conf = conf.databases.__getattr__(database)

            else:
                conf = conf.caches.__getattr__(database)

        except AttributeError:
            conf = None

        return conf

    def engine(self, config_dsn, cache=False):
        delegate = {}
        conf = self._parse_config(config_dsn, delegate, cache=cache)

        if not conf:
            return None

        if not delegate['key'] in ModelSingleton().engines:
            ModelSingleton._lock.acquire()

            if conf.driver == 'memory':
                identifier = delegate['database']

                if conf.identifier:
                    identifier = conf.identifier

                app_path = self.ini.server.application_path
                path = '%s/resource/database/sqlite/cache_%s_%s.db' \
                       % (app_path, delegate['config'], identifier)

                connection_args = {'check_same_thread': False}
                connection_url = 'sqlite:///%s' % path

            elif conf.driver == 'sqlite':
                identifier = delegate['database']

                if conf.identifier:
                    identifier = conf.identifier

                if conf.path:
                    path = conf.path
                else:
                    app_path = self.ini.server.application_path
                    path = '%s/resource/database/sqlite/%s_%s.db' \
                           % (app_path, delegate['config'], identifier)

                connection_args = {'check_same_thread': False}
                connection_url = 'sqlite:///%s' % path

            else:
                connection_args = {}
                connection_url_args = {}

                if conf.charset:
                    connection_url_args['charset'] = conf.charset

                for e in conf.kwargs or []:
                    connection_url_args[e] = getattr(conf.kwargs, e)

                connection_url = '%s://%s:%s@%s:%s/%s' \
                                 % (conf.driver, conf.user, conf.password, conf.host, conf.port, conf.database)
                connection_url = self.helper.web.url.build(connection_url, connection_url_args)

            params = {
                'convert_unicode': conf.convert_unicode if conf.convert_unicode is not None else True,
                'echo': conf.echo if conf.echo is not None else False,
                'echo_pool': conf.echo_pool if conf.echo_pool is not None else False,
                'pool_size': conf.pool_size if conf.pool_size is not None else 16,
                'poolclass': QueuePool,
                'pool_recycle': conf.pool_recycle if conf.pool_recycle is not None else 3600,
                'max_overflow': conf.max_overflow if conf.max_overflow is not None else -1,
                'pool_timeout': conf.pool_timeout if conf.pool_timeout is not None else 30,
                'connect_args': connection_args
            }

            if conf.isolation_level is not None:
                params['isolation_level'] = conf.isolation_level

            ModelSingleton().engines[delegate['key']] = create_engine(connection_url, **params)

            ModelSingleton._lock.release()

        return ModelSingleton().engines[delegate['key']]

    def getconn(self, config_dsn, cache=False):
        engine = ModelSingleton().engine(config_dsn, cache=cache)

        if not engine:
            return None

        return engine.connect()

    def begin(self, dsn_or_conn, cache=False):
        config_dsn = dsn_or_conn if isinstance(dsn_or_conn, (str, InValueModelConfig)) else None
        connection = self.getconn(config_dsn, cache=cache) if config_dsn else dsn_or_conn
        transaction = connection.begin()

        return ModelProxy(connection, transaction)

    def commit(self, proxy, return_connection=True):
        result = proxy.transaction.commit()

        if return_connection:
            proxy.connection.close()

        return result

    def rollback(self, proxy, return_connection=True):
        result = proxy.transaction.rollback()

        if return_connection:
            proxy.connection.close()

        return result

    def close(self, proxy):
        return proxy.connection.close()

    def _bind_helper(self, conn, sql, bind):
        """
          tuple style binding -> dict style binding

        if conn and \
                getattr(conn, 'engine', None) and \
                getattr(getattr(conn, 'engine'), 'url', None) and \
                getattr(getattr(conn, 'engine'), 'url').drivername == 'sqlite':
            if sql.find('%s') != -1 and isinstance(bind, (tuple, list)):
                param = tuple([':param_%s' % k for k in range(len(bind))])
                bind = dict([('param_%s' % k, bind[k]) for k in range(len(bind))])
                sql = sql % param
        """

        return sql, bind

    def execute(self, sql, bind=None, dsn_or_conn=None, cache=False):
        config_dsn = dsn_or_conn if isinstance(dsn_or_conn, (str, InValueModelConfig)) else None
        conn = self.getconn(config_dsn, cache=cache) if config_dsn else dsn_or_conn
        conn = conn.connection if isinstance(conn, ModelProxy) else conn

        sql, bind = self._bind_helper(conn, sql, bind)

        if isinstance(bind, dict):
            result = conn.execute(sql, **bind)
        elif isinstance(bind, list):
            result = conn.execute(sql, *bind)
        elif bind is not None:
            result = conn.execute(sql, (bind, ))
        else:
            result = conn.execute(sql)

        if config_dsn:
            conn.close()

        return result

    def row(self, sql, bind=None, dsn_or_conn=None, cache=False, *args, **kwargs):
        config_dsn = dsn_or_conn if isinstance(dsn_or_conn, (str, InValueModelConfig)) else None
        conn = self.getconn(config_dsn, cache=cache) if config_dsn else dsn_or_conn
        result = self.execute(sql, bind, dsn_or_conn=conn, cache=cache)
        to_dict = True if 'to_dict' in kwargs else False

        row = None

        for r in result:
            row = r if not to_dict else dict(r.items())
            break

        if config_dsn:
            conn.close()

        return row

    def rows(self, sql, bind=None, dsn_or_conn=None, cache=False, *args, **kwargs):
        config_dsn = dsn_or_conn if isinstance(dsn_or_conn, (str, InValueModelConfig)) else None
        conn = self.getconn(config_dsn, cache=cache) if config_dsn else dsn_or_conn
        result = self.execute(sql, bind, dsn_or_conn=conn, cache=cache)
        to_dict = True if 'to_dict' in kwargs else False

        rows = []

        for r in result:
            rows.append(r if not to_dict else dict(r.items()))

        if config_dsn:
            conn.close()

        return rows

    def scalar(self, sql, bind=None, dsn_or_conn=None, cache=False):
        config_dsn = dsn_or_conn if isinstance(dsn_or_conn, (str, InValueModelConfig)) else None
        conn = self.getconn(config_dsn, cache=cache) if config_dsn else dsn_or_conn
        result = self.execute(sql, bind, dsn_or_conn=conn, cache=cache)
        value = result.scalar() if result else None

        if config_dsn:
            conn.close()

        return value


class ModelProxy(object):
    def __init__(self, connection, transaction):
        self._connection = connection
        self._transaction = transaction

    @property
    def connection(self):
        return self._connection

    @property
    def transaction(self):
        return self._transaction

    def commit(self, return_connection=True):
        return ModelSingleton().commit(self, return_connection)

    def rollback(self, return_connection=True):
        return ModelSingleton().rollback(self, return_connection)

    def close(self):
        return ModelSingleton().close(self)

    def execute(self, sql, bind=None, dsn_or_conn=None):
        return ModelSingleton().execute(sql, bind, dsn_or_conn or self)

    def row(self, sql, bind=None, dsn_or_conn=None):
        return ModelSingleton().row(sql, bind, dsn_or_conn or self)

    def rows(self, sql, bind=None, dsn_or_conn=None):
        return ModelSingleton().rows(sql, bind, dsn_or_conn or self)

    def scalar(self, sql, bind=None, dsn_or_conn=None):
        return ModelSingleton().scalar(sql, bind, dsn_or_conn or self)


class Model(dpEngine, dpLoader):
    @staticmethod
    def caching(*args, **kwargs):
        return ModelSingleton().cache.decorator(*args, **kwargs)

    @staticmethod
    def clear_cached(method, *args, **kwargs):
        return ModelSingleton().cache.clear(method, *args, **kwargs)

    @staticmethod
    def renew_cached(method, *args, **kwargs):
        return ModelSingleton().cache.renew(method, *args, **kwargs)

    @staticmethod
    def run_alone(*args, **kwargs):
        return ModelSingleton().cache.lock_decorator(*args, **kwargs)

    def _getconn(self, config_dsn):
        return ModelSingleton().getconn(config_dsn)

    def begin(self, dsn_or_conn):
        return ModelSingleton().begin(dsn_or_conn)

    def commit(self, proxy, return_connection=True):
        return ModelSingleton().commit(proxy, return_connection)

    def rollback(self, proxy, return_connection=True):
        return ModelSingleton().rollback(proxy, return_connection)

    def close(self, proxy):
        return ModelSingleton().close(proxy)

    def execute(self, sql, bind=None, dsn_or_conn=None):
        return ModelSingleton().execute(sql, bind, dsn_or_conn)

    def row(self, sql, bind=None, dsn_or_conn=None, *args, **kwargs):
        return ModelSingleton().row(sql, bind, dsn_or_conn, *args, **kwargs)

    def rows(self, sql, bind=None, dsn_or_conn=None, *args, **kwargs):
        return ModelSingleton().rows(sql, bind, dsn_or_conn, *args, **kwargs)

    def scalar(self, sql, bind=None, dsn_or_conn=None):
        return ModelSingleton().scalar(sql, bind, dsn_or_conn)



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class CountryHelper(dpHelper):
    @property
    def south_korea(self):
        return 'ko_KR'

    @property
    def available_countries(self):
        return self.south_korea,



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class RandomHelper(dpHelper):
    def int(self, length):
        return self.helper.misc.random.numeric(length)

    def range(self, minval, maxval):
        if not self.helper.misc.type.check.numeric(minval) or not self.helper.misc.type.check.numeric(maxval):
            return False

        return self.helper.misc.random.randint(minval, maxval)



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper

try:
    # py 2.x
    import HTMLParser
    html_parser = HTMLParser.HTMLParser()

except:
    # py 3.4-
    try:
        import html.parser
        html_parser = html.parser.HTMLParser()
    except:
        # py 3.4+
        import html as html_parser

try:
    import htmltag
except:
    htmltag = None

import re

from bs4 import BeautifulSoup


re_html_tag = re.compile("(?i)<\/?\w+((\s+\w+(\s*=\s*(?:\".*?\"|'.*?'|[^'\">\s]+))?)+\s*|\s*)\/?>")
re_on_attrs = re.compile('.*\s+(on[a-z]+\s*=).*')


class HtmlHelper(dpHelper):
    def validate(self, s):
        s_id = '_____dp_s_xss_____'
        s = '<div id="%s">%s</div>' % (s_id, s)
        s = BeautifulSoup(s, 'lxml')

        s = str(s.find(id=s_id))
        return s[s.find('>')+1:s.rfind('<')]

    def strip_xss(self, s, whitelist=None):
        if whitelist is None:
            whitelist = (
                'a', 'abbr', 'aside', 'audio', 'bdi', 'bdo', 'blockquote', 'canvas',
                'caption', 'code', 'col', 'colgroup', 'data', 'dd', 'del',
                'details', 'div', 'dl', 'dt', 'em', 'figcaption', 'figure', 'h1',
                'h2', 'h3', 'h4', 'h5', 'h6', 'hr', 'i', 'img', 'ins', 'kbd', 'li',
                'mark', 'ol', 'p', 'pre', 'q', 'rp', 'rt', 'ruby', 's', 'samp',
                'small', 'source', 'span', 'strong', 'sub', 'summary', 'sup',
                'table', 'td', 'th', 'time', 'tr', 'track', 'u', 'ul', 'var',
                'video', 'wbr', 'b', 'br', 'site', 'font')

        elif not whitelist:
            whitelist = None

        s_id = '_____dp_s_xss_____'
        s = '<div id="%s">%s</div>' % (s_id, s)
        s = BeautifulSoup(s, 'lxml')

        self._strip_xss(s.find(id=s_id), whitelist)

        s = str(s.find(id=s_id))
        return s[s.find('>')+1:s.rfind('<')]

    def _strip_xss(self, elements, whitelist):
        for e in elements.find_all():
            found = False

            if whitelist and e.name not in whitelist:
                found = True
            elif e.attrs:
                for ea_k, ea_v in e.attrs.items():
                    s_tag = ea_k.strip().lower()

                    if s_tag.startswith('on') or s_tag.startswith('seeksegmenttime') or s_tag.startswith('fscommand'):
                        found = True
                        break

                    s_val = ea_v.strip()

                    for xe in self.helper.string.whitespace:
                        s_val = s_val.replace(xe, '')

                    if s_val.find('javascript:') != -1 or s_val.find('vbscript:') != -1:
                        found = True
                        break

            if found:
                e.unwrap()

            self._strip_xss(e, whitelist)

    def strip_tags(self, text):
        return re.sub('<[^<]+?>', '', text)

    def unescape(self, text):
        return html_parser.unescape(text)

    def escape(self, s, quote=False):
        s = s.replace("&", "&amp;")
        s = s.replace("<", "&lt;")
        s = s.replace(">", "&gt;")

        if quote:
            s = s.replace('"', "&quot;")

        return s



# -*- coding: utf-8 -*-


import boto3

from dp_tornado.engine.helper import Helper as dpHelper


class DynamodbHelper(dpHelper):
    def client(
            self,
            access_key_id,
            secret_access_key,
            region_name):
        return boto3.client(
            service_name='dynamodb',
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            region_name=region_name)

    def resource(
            self,
            access_key_id,
            secret_access_key,
            region_name):
        return boto3.resource(
            service_name='dynamodb',
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            region_name=region_name)





# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper


class JsonHelper(dpHelper):
    def stringify(self, obj, beautify=False, raise_exception=False):
        """Alias of helper.string.serialization.json.stringify"""

        return self.helper.string.serialization.json.stringify(
            obj=obj,
            beautify=beautify,
            raise_exception=raise_exception)

    def parse(self, text, encoding='utf8', raise_exception=False):
        """Alias of helper.string.serialization.json.parse"""

        return self.helper.string.serialization.json.parse(
            text=text,
            encoding=encoding,
            raise_exception=raise_exception)



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class TypeHelper(dpHelper):
    @property
    def int(self):
        return int

    @property
    def long(self):
        if self.helper.misc.system.py_version <= 2:
            return long
        else:
            return int

    @property
    def float(self):
        return float

    @property
    def bool(self):
        return bool

    @property
    def numeric(self):
        if self.helper.misc.system.py_version <= 2:
            return int, long, float
        else:
            return int, float

    @property
    def string(self):
        if self.helper.misc.system.py_version <= 2:
            return basestring,
        else:
            return str,

    @property
    def array(self):
        return list, tuple

    @property
    def dict(self):
        return dict,



# -*- coding: utf-8 -*-


import tempfile

try:
    from PIL import Image
except ImportError:
    Image = None

try:
    from PIL import ImageDraw
except ImportError:
    ImageDraw = None

try:
    from PIL import ImageOps
except ImportError:
    ImageOps = None

from dp_tornado.engine.helper import Helper as dpHelper


class ImageHelper(dpHelper):
    def compare(self, i1, i2, error=0):
        i1 = self.load(i1)
        i2 = self.load(i2)

        if not i1 or not i2:
            return None

        s1 = i1.size
        s2 = i2.size

        if s1[0] != s2[0] or s2[1] != s2[1]:
            print('size ne,', s1, s2)
            return False

        i1 = i1.load()
        i2 = i2.load()

        for i in range(s1[0]):
            for j in range(s1[1]):
                if i1[i, j] != i2[i, j]:
                    if error:
                        for k in range(len(i1[i, j])):
                            if abs(i1[i, j][k] - i2[i, j][k]) > error:
                                print('pixel ne,', i1[i, j], i2[i, j], abs(i1[i, j][k] - i2[i, j][k]), error)
                                return False
                    else:
                        return False

        return True

    def load(self, src):
        tmp = None

        try:
            if isinstance(src, Image.Image):
                return src

            elif self.helper.web.url.validate(src):
                code, res = self.helper.web.http.get.raw(src)

                if code != 200:
                    return False

                tmp = tempfile.NamedTemporaryFile(delete=False)
                tmp.write(res)
                tmp.close()

                tmp = tmp.name

            else:
                tmp = None

            img = Image.open(tmp if tmp else src)

            if not img:
                raise Exception('The specified image is invalid.')

            return img

        except Exception as e:
            self.logging.exception(e)

            return False

        finally:
            if tmp:
                self.helper.io.file.remove(tmp)

    def execute(self, src, fn, **kwargs):
        img = self.load(src)

        if not img:
            return False

        try:
            return fn(img, **kwargs)

        except Exception as e:
            self.logging.exception(e)

            return False

    def size(self, src, **o_kwargs):
        def fn(img, **kwargs):
            if not img:
                return -1, -1

            return img.width, img.height

        return self.execute(src, fn, **o_kwargs)

    def crop(self, src, **o_kwargs):
        def fn(img, **kwargs):
            crop = kwargs['crop'] if 'crop' in kwargs else None

            if not crop:
                return img

            e_top = 0
            e_left = 0
            e_right = 0
            e_bottom = 0

            if self.helper.misc.type.check.string(crop):
                crop = crop.split(',')
                crop = [int(e.strip()) for e in crop]

            if self.helper.misc.type.check.numeric(crop):
                e_top = e_left = e_right = e_bottom = crop

            elif isinstance(crop, (tuple, list)):
                if len(crop) == 1:
                    e_top = e_left = e_right = e_bottom = crop[0]

                elif len(crop) == 2:
                    e_top = e_bottom = crop[0]
                    e_left = e_right = crop[1]

                elif len(crop) == 4:
                    e_top = crop[0]
                    e_right = crop[1]
                    e_bottom = crop[2]
                    e_left = crop[3]

            img = img.crop((e_left, e_top, img.size[0] - e_right, img.size[1] - e_bottom))

            return img

        return self.execute(src, fn, **o_kwargs)

    def border(self, src, **o_kwargs):
        def fn(img, **kwargs):
            border = int(kwargs['border']) if 'border' in kwargs else 0
            border_color = kwargs['border_color'] if 'border_color' in kwargs else '#000000'

            if not border:
                return img

            if '_org' in kwargs and 'radius' in kwargs and kwargs['radius']:
                return img

            if img.mode == 'P':
                img = img.convert('RGBA')

            img = ImageOps.expand(img, border=(border, border), fill=border_color)

            return img

        return self.execute(src, fn, **o_kwargs)

    def radius(self, src, **o_kwargs):
        def fn(img, **kwargs):
            radius = int(kwargs['radius'] or 0) if 'radius' in kwargs else None
            border = int(kwargs['border']) if 'border' in kwargs else 0
            border_color = kwargs['border_color'] if 'border_color' in kwargs else '#000000'

            if not radius:
                return img
            elif '__radius_processed__' in img.__dict__:
                return img

            img = img.convert('RGBA')
            img.putalpha(self._rounded_mask(img.size, min(radius, *img.size)))

            # Border with radius
            if border:
                img_o = img
                bordered_size = (img.size[0] + (border * 2), img.size[1] + (border * 2))
                img = Image.new('RGBA', bordered_size, border_color)
                img.putalpha(self._rounded_mask(bordered_size, min(radius, *bordered_size)))
                img.paste(img_o, (border, border), img_o)

            img.__dict__['__radius_processed__'] = True

            return img

        return self.execute(src, fn, **o_kwargs)

    def colorize(self, src, **o_kwargs):
        def fn(img, **kwargs):
            colorize = kwargs['colorize'] if 'colorize' in kwargs else None

            if not colorize:
                return img

            img = img.convert('RGBA')
            r, g, b, a = img.split()
            gray = ImageOps.grayscale(img)
            result = ImageOps.colorize(gray, colorize, (255, 255, 255, 0))
            result.putalpha(a)
            img = result

            return img

        return self.execute(src, fn, **o_kwargs)

    def resize(self, src, **o_kwargs):
        def fn(img, **kwargs):
            size = kwargs['size'] if 'size' in kwargs else None
            mode = kwargs['mode'] if 'mode' in kwargs else None
            scale = int(kwargs['scale']) if 'scale' in kwargs else 1
            limit = True if 'limit' in kwargs and kwargs['limit'] else False
            background = kwargs['background'] if 'background' in kwargs else None
            border = int(kwargs['border']) if 'border' in kwargs else 0

            if not size:
                return img

            width_new, height_new = size
            width_origin, height_origin = img.size

            if scale > 1:
                if limit:
                    scale_max_width = float(width_origin) / float(width_new)
                    scale_max_height = float(height_origin) / float(height_new)

                    scale_max = min(scale, scale_max_width, scale_max_height)
                else:
                    scale_max = scale

                if scale_max > 1:
                    width_new = int(width_new * scale_max)
                    height_new = int(height_new * scale_max)

            if not width_new:
                width_new = width_origin * height_new / height_origin
                mode = self.helper.io.image.mode.resize

            if not height_new:
                height_new = height_origin * width_new / width_origin
                mode = self.helper.io.image.mode.resize

            if border:
                width_new -= border * 2
                height_new -= border * 2

            if not mode:
                mode = self.helper.io.image.mode.resize

            if mode not in self.helper.io.image.mode.modes:
                raise Exception('The specified mode is not supported.')

            # Image Resizing
            if mode == self.helper.io.image.mode.center:
                width_calc = width_new
                height_calc = height_origin * width_calc / width_origin

                if height_calc > height_new:
                    height_calc = height_new
                    width_calc = width_origin * height_calc / height_origin

                img = img.resize((int(width_calc), int(height_calc)), Image.ANTIALIAS)
                img = self.radius(img, **kwargs)

                img = ImageOps.expand(
                    img, border=(int((width_new - width_calc) / 2), int((height_new - height_calc) / 2)), fill=background)

                radius = int(kwargs['radius'] or 0) if 'radius' in kwargs else None

                if radius:
                    img_empty = Image.new('RGB', (width_new, height_new), background)
                    img_empty.paste(img, (0, 0), img)
                    img = img_empty

                img.__dict__['__radius_processed__'] = True

            elif mode == self.helper.io.image.mode.fill:
                ratio_origin = float(width_origin) / float(height_origin)
                ratio_new = float(width_new) / float(height_new)

                if ratio_origin > ratio_new:
                    tw = int(round(height_new * ratio_origin))
                    img = img.resize((tw, height_new), Image.ANTIALIAS)
                    left = int(round((tw - width_new) / 2.0))
                    img = img.crop((left, 0, left + width_new, height_new))

                elif ratio_origin < ratio_new:
                    th = int(round(width_new / ratio_origin))
                    img = img.resize((width_new, th), Image.ANTIALIAS)
                    top = int(round((th - height_new) / 2.0))
                    img = img.crop((0, top, width_new, top + height_new))

                else:
                    img = img.resize((width_new, height_new), Image.ANTIALIAS)

            elif mode == self.helper.io.image.mode.resize:
                if width_new > width_origin or height_new > height_origin:
                    width_new = width_origin
                    height_new = height_origin

                img = img.resize((width_new, height_new), Image.ANTIALIAS)

            return img

        return self.execute(src, fn, **o_kwargs)

    def save(self, src, **o_kwargs):
        def fn(img, **kwargs):
            ext = kwargs['format'] if 'format' in kwargs else None
            dest = kwargs['dest'] if 'dest' in kwargs else None

            if not dest:
                return None

            if not ext and self.helper.misc.type.check.string(dest):
                ext = self.helper.io.path.ext(dest, dot='').lower()

            if not ext and self.helper.misc.type.check.string(src):
                ext = self.helper.io.path.ext(src, dot='').lower()

            if not ext and '_org' in kwargs and kwargs['_org'] and self.helper.misc.type.check.string(kwargs['_org']):
                ext = self.helper.io.path.ext(kwargs['_org'], dot='').lower()

            if dest == 's3':
                # TODO
                return False

            # PNG : 1, L, P, RGB, RGBA
            # GIF : L, P
            # JPEG : L, RGB, CMYK

            # Invalid format specified
            if ext in ('jpg', 'jpeg') and img.mode == 'P':  # JPEG does not support P mode
                img = img.convert('RGBA')

            img.save(dest, format=ext, quality=100)

            return True

        return self.execute(src, fn, **o_kwargs)

    def manipulate(self, src, **kwargs):
        kwargs['_org'] = src

        try:
            img = self.load(src)

            # Crop
            img = self.crop(img, **kwargs)

            if not img:
                return False

            # Resize
            img = self.resize(img, **kwargs)

            if not img:
                return False

            # Radius
            img = self.radius(img, **kwargs)

            if not img:
                return False

            # Border
            img = self.border(img, **kwargs)

            if not img:
                return False

            # Colorize
            img = self.colorize(img, **kwargs)

            if not img:
                return False

            # Save
            saved = self.save(img, **kwargs)

            if saved is None:
                return img
            elif saved is False:
                return False

            return True

        except Exception as e:
            self.logging.exception(e)

            return False

    def _rounded_mask(self, size, radius, factor=2):
        width, height = size
        radius = min(radius, *size)

        mask = Image.new("L", (width * factor, height * factor))
        mask_draw = ImageDraw.Draw(mask)
        mask_draw.ellipse((0, 0, radius * factor, radius * factor), fill='#FFFFFF')

        flip = mask.transpose(Image.FLIP_LEFT_RIGHT)
        mask.paste(flip, (0, 0), flip)
        mask_draw.rectangle((radius * factor / 2, 0, (width - (radius / 2)) * factor, radius * factor), fill='#FFFFFF')
        flip = mask.transpose(Image.FLIP_TOP_BOTTOM)
        mask.paste(flip, (0, 0), flip)
        mask_draw.rectangle((0, radius * factor / 2, width * factor, (height - (radius / 2)) * factor), fill='#FFFFFF')
        mask = mask.resize((width, height), Image.ANTIALIAS)

        return mask



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class PhoneHelper(dpHelper):
    def validate(self, p, locale=None):
        locale = locale or self.helper.locale.country.south_korea

        if locale == self.helper.locale.country.south_korea:
            return self.helper.locale.korea.validate_phone_number(p)
        else:
            raise Exception('The specified locale is not supported.')



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper

unicode_type = type(u'')


class CastHelper(dpHelper):
    def string(self, s, preserve_none=True):
        if s is None:
            return s if preserve_none else ''

        if not self.helper.misc.type.check.string(s):
            s = str(s)

        if type(s) == unicode_type:
            if self.helper.misc.system.py_version <= 2:
                return s.encode('UTF-8')
            else:
                return s
        else:
            return s

    def unicode(self, s, preserve_none=True):
        if s is None:
            return s if preserve_none else u''

        if not self.helper.misc.type.check.string(s):
            s = str(s)

        if type(s) != unicode_type:
            return s.decode('UTF-8')
        else:
            return s



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper
from validate_email import validate_email


class EmailHelper(dpHelper):
    def validate(self, e):
        return validate_email(e)



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper
from Crypto.Cipher import AES


class CryptoHelper(dpHelper):
    def encrypt(self, plain, randomized=False, expire_in=0, key=None, raw=False, encode=True, pad=True):
        if raw:
            return self._encrypt(plain, key=key, encode=encode, pad=pad)

        payload = {
            'p': plain
        }

        if randomized and not expire_in:
            payload['r'] = self.helper.datetime.timestamp.now(ms=True)

        if expire_in and expire_in > 0:
            payload['e'] = self.helper.datetime.timestamp.now(ms=True) + (expire_in * 1000)

        payload = self.helper.string.serialization.serialize(payload, method='json')
        return self._encrypt(payload, key=key)

    def decrypt(self, encrypted, key=None, raw=False, encode=True, pad=True):
        decrypted = self._decrypt(encrypted, key=key, encode=encode, pad=pad)

        if raw:
            return decrypted

        payload = self.helper.string.serialization.deserialize(decrypted, method='json')

        if not payload:
            return None

        if 'e' in payload and self.helper.datetime.timestamp.now(ms=True) > payload['e']:
            return False

        return payload['p']

    def _pad(self, s):
        if self.helper.misc.system.py_version >= 3:
            return s + (16 - len(s) % 16) * chr(16 - len(s) % 16).encode('utf8')
        else:
            return s + (16 - len(s) % 16) * chr(16 - len(s) % 16)

    def _unpad(self, s):
        return s[:-ord(s[len(s)-1:])]

    def _key_and_iv(self, key):
        if isinstance(key, (tuple, list)):
            return key[0], key[1]

        if not key:
            key = getattr(self, '__crypto_default_key__', None)

            if not key:
                key = self.ini.crypto.key
                setattr(self, '__crypto_default_key__', key)

        r_key = key * max(1, int(self.helper.numeric.math.ceil(32.0 / (len(key) * 1.0))))
        return r_key[0:16], r_key[16:32]

    def _encrypt(self, plain, key=None, encode=True, pad=True):
        key, iv = self._key_and_iv(key)

        if self.helper.misc.system.py_version >= 3:
            plain = plain.encode('utf8')

        encrypted = AES.new(key, AES.MODE_CBC, iv).encrypt(self._pad(plain) if pad else plain)
        return self.helper.security.crypto.encoding.base64_encode(encrypted, raw=True) if encode else encrypted

    def _decrypt(self, encrypted, key=None, encode=True, pad=True):
        key, iv = self._key_and_iv(key)
        encrypted = self.helper.security.crypto.encoding.base64_decode(encrypted, raw=True) if encode else encrypted
        decrypted = AES.new(key, AES.MODE_CBC, iv).decrypt(encrypted)
        decrypted = self._unpad(decrypted) if pad else decrypted

        if self.helper.misc.system.py_version >= 3:
            decrypted = decrypted.decode('utf8')

        return decrypted



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class CheckHelper(dpHelper):
    def exist_repeated_text(self, s, criteria=3):
        if not self.helper.misc.type.check.string(s):
            return False

        k = s[0]
        n = 0

        for c in s:
            if c == k:
                n += 1

                if n >= criteria:
                    return True

            else:
                k = c
                n = 1

        return False

    def alphanumericpunc(self, s, add=None):
        return self._check(
            s, self.helper.string.digits + self.helper.string.ascii_letters + self.helper.string.punctuation, add)

    def alphanumeric(self, s, add=None):
        return self._check(s, self.helper.string.digits + self.helper.string.ascii_letters, add)

    def alphabet(self, s, add=None):
        return self._check(s, self.helper.string.ascii_letters, add)

    def numeric(self, s, add=None):
        return self._check(s, self.helper.string.digits, add)

    def _check(self, s, criteria, add):
        if not self.helper.misc.type.check.string(s):
            return False

        add = add if add else ''
        v = any(char not in set(c for c in criteria + add) for char in s)
        return True if not v else False



# -*- coding: utf-8 -*-
"""A view can be any output representation of information, such as a chart or a diagram. `<Wikipedia>
<https://en.wikipedia.org/wiki/Model–view–controller>`_

Here is a view example:

.. testcode::

    from dp_tornado.engine.controller import Controller

    class FooBarController(Controller):
        def get(self):
            self.render('index.html')
"""

from dp_tornado.engine.singleton import Singleton as dpSingleton
from dp_tornado.engine.plugin.static import StaticURL
from dp_tornado.engine.plugin.pagination import Pagination
from dp_tornado.engine.plugin import ui_methods
from dp_tornado.engine.engine import Engine as dpEngine

from tornado import template

import os

engine = dpEngine()


class Dummy(object):
    pass


class View(dpSingleton):
    @staticmethod
    def _ui_modules():
        if getattr(View, '__ui_modules', None) is not None:
            return getattr(View, '__ui_modules')

        o = {
            'Static': StaticURL,
            'Pagination': Pagination
        }

        setattr(View, '__ui_modules', o)

        return o

    @staticmethod
    def _ui_methods():
        if getattr(View, '__ui_methods', None) is not None:
            return getattr(View, '__ui_methods')

        o = {}

        for e in dir(ui_methods):
            if not e.startswith("_") and e[0].lower() == e[0]:
                o[e] = getattr(ui_methods, e)

        setattr(View, '__ui_methods', o)

        return o

    @staticmethod
    def _ui_namespaces(controller):
        o = {
            'c': lambda *args, **kwargs: controller,
            '_tt_modules': Dummy()
        }

        for k, v in View._ui_modules().items():
            setattr(o['_tt_modules'], k, v(controller).render)

        for k, v in View._ui_methods().items():
            o[k] = View._ui_method(controller, v)

        return o

    @staticmethod
    def _ui_method(controller, method):
        return lambda *args, **kwargs: method(controller, *args, **kwargs)

    @staticmethod
    def _loader():
        if getattr(View, '__loader', None) is not None:
            return getattr(View, '__loader')

        o = template.Loader(os.path.join(engine.ini.server.application_path, 'view'))
        setattr(View, '__loader', o)

        return o

    @staticmethod
    def render(controller, template_name, kwargs=None):
        controller.finish(View.render_string(template_name, kwargs, encode=False))

    @staticmethod
    def render_string(controller, template_name, kwargs=None, encode=True):
        t = View._loader().load(template_name)
        n = View._ui_namespaces(controller)
        n.update(kwargs if kwargs else {})
        t = t.generate(**n)

        if encode is True:
            if engine.helper.misc.system.py_version <= 2:
                return engine.helper.string.cast.string(t)
            else:
                return str(t, 'UTF-8')

        return t

    @staticmethod
    def write(controller, chunk):
        controller.parent.write(chunk)

    @staticmethod
    def finish(controller, chunk=None):
        controller.parent.finish(chunk)



# -*- coding: utf-8 -*-


import importlib

from .engine import EngineSingleton as dpEnginSingleton


engine = dpEnginSingleton()


class M17n(object):
    _prepared = {}

    def set(self, controller, m17n_lang):
        if not m17n_lang:
            controller.clear_cookie('__m17n__')

        if m17n_lang not in engine.ini.server.m17n:
            raise Exception('Not allowed multilingual language.')

        return controller.set_cookie('__m17n__', m17n_lang, expires_days=365*10)

    def get(self, m17n_lang=None):
        if m17n_lang and isinstance(m17n_lang, (list, tuple)):
            m17n_lang = m17n_lang[0]
        elif not m17n_lang or m17n_lang not in engine.ini.server.m17n:
            m17n_lang = engine.ini.server.m17n[0]

        if m17n_lang in M17n._prepared:
            return M17n._prepared[m17n_lang]

        M17n._prepared[m17n_lang] = M17nLoader(m17n_lang)
        return M17n._prepared[m17n_lang]


class M17nLoader(object):
    def __init__(self, m17n_lang, path_prefix=None):
        self.__dict__['_m17n_lang'] = m17n_lang
        self.__dict__['_path_prefix'] = path_prefix or 'm17n'
        self.__dict__['_prepared'] = {}

    def m17n_cache(self, traverse):
        if traverse in self.__dict__['_prepared']:
            return self.__dict__['_prepared'][traverse]

        traversed = eval('self.%s' % traverse)
        self.__dict__['_prepared'][traverse] = traversed
        return self.__dict__['_prepared'][traverse]

    def __getattr__(self, name):
        try:
            attr = self.__getattribute__(name)
        except AttributeError:
            attr = None

        if attr:
            return attr

        if name in self.__dict__:
            return self.__dict__[name]

        if name == '_':
            return self.m17n_cache

        if name in self.__dict__['_prepared']:
            return self.__dict__['_prepared'][name]
        
        try:
            path_prefix = '%s.%s' % (self.__dict__['_path_prefix'], name)
            importlib.import_module(path_prefix)

            self.__dict__['_prepared'][name] = M17nLoader(self.__dict__['_m17n_lang'], path_prefix=path_prefix)
            return self.__dict__['_prepared'][name]

        except ImportError:
            try:
                path_prefix = '%s.%s' % (self.__dict__['_path_prefix'], self.__dict__['_m17n_lang'])
                module = importlib.import_module(path_prefix)
                m17n_obj = getattr(module, 'M17n')

                if m17n_obj and hasattr(m17n_obj, name):
                    self.__dict__['_prepared'][name] = getattr(m17n_obj, name)
                    return self.__dict__['_prepared'][name]

            except Exception:
                pass

            return '{{ %s.%s }}' % (self.__dict__['_path_prefix'], name)



# -*- coding: utf-8 -*-


from dp_tornado.engine.engine import Engine as dpEngine
from dp_tornado.engine.bootstrap import Bootstrap as EngineBootstrap
from dp_tornado.engine.testing import Testing as dpTesting
from dp_tornado.version import __version_info__


class CliHandler(dpEngine):
    def __init__(self):
        import argparse

        parser = argparse.ArgumentParser()

        args = [
            ['action', {'nargs': 1}],
            ['options', {'nargs': '*'}],
            ['--identifier', {'help': 'Identifier'}],
            ['--dryrun', {'help': 'Dryrun, If this value specified `yes` then enabled.'}],
            ['--template', {'help': 'Template Name', 'default': 'helloworld'}],
            ['--logging', {'help': 'Logging'}],
            ['--path', {'help': 'App Path'}],
            ['--port', {'help': 'Binding port', 'type': int}]
        ]

        for e in args:
            parser.add_argument(e[0], **e[1])

        self.parser = parser
        self.args, self.args_unknown = parser.parse_known_args()
        self.cwd = self.helper.io.path.cwd()

    def main(self):
        self.logging.info('------------------------')
        self.logging.info('* dp for Python v%s' % '.'.join([str(e) for e in __version_info__]))
        self.logging.info('------------------------')

        for e in self.args.options:
            if not self.args.path:
                path = self.helper.io.path.join(self.cwd, e)

                if self.helper.io.path.is_file(path) or \
                        (self.helper.io.path.mkdir(path) and self.helper.io.path.is_dir(path)):
                    self.args.path = e

        if self.args.action and self.args.action[0] == 'init':
            self.command_init()

        elif self.args.action and self.args.action[0] == 'run':
            self.command_run()

        elif self.args.action and self.args.action[0] == 'test':
            self.command_test()

        else:
            self.logging.info('* dp4p finished, unrecognized action.')

            import sys
            self.logging.info('%s' % sys.argv)

            return exit(1)

        return exit(0)

    def command_init(self):
        installable = True
        init_path = self._pathified()

        self.logging.info('* Initializing app .. %s' % init_path)

        if self.helper.io.path.is_dir(init_path):
            browse = self.helper.io.path.browse(init_path)
            browse = [e for e in browse if not self.helper.io.path.split(e)[1].startswith('.')]

            if len(browse) > 0:
                status = 'Not Empty'
                installable = False
            else:
                status = 'Empty'

        elif self.helper.io.path.is_file(init_path):
            status = 'File'
            installable = False

        else:
            self.helper.io.path.mkdir(init_path)

            if not self.helper.io.path.is_dir(init_path):
                status = 'Permission Denied'
                installable = False
            else:
                status = 'Empty'

        if not installable:
            self.logging.info('* Initialization failed, %s' % status)
            return exit(1)

        engine_path = self.helper.io.path.dirname(__file__)
        application_path = init_path

        # template initialization.
        if not EngineBootstrap.init_template(
                engine_path=engine_path, application_path=application_path, template_name=self.args.template):
            self.logging.info('* Initialization failed.')
            return exit(1)

        self.logging.info('* Initialization succeed.')

    def command_run(self):
        init_path = self._pathified()

        init_py = init_path
        executable = True

        if self.helper.io.path.is_dir(init_py):
            init_py = '%s/__init__.py' % init_py

        self.logging.info('* Running app .. %s' % init_py)

        if not self.helper.io.path.is_file(init_py):
            executable = False

        if not executable:
            self.logging.info('* Running failed, Not executable path.')
            return exit(1)

        app_module = self._append_sys_path(init_path)

        import sys

        __import__(app_module)
        app = sys.modules[app_module] if app_module in sys.modules else None
        app_run = getattr(app, 'run', None) if app else None

        if not app_run:
            self.logging.info('* Running failed, Invalid app.')
            return exit(1)

        for i in range(len(self.args.options) + 1):
            sys.argv.pop(1)

        try:
            app_run(self)

        except KeyboardInterrupt:
            pass

        except Exception as e:
            self.logging.exception(e)

    def command_test(self):
        tester = dpTesting(self._append_sys_path(), self._pathified())

        if not tester.traverse():
            self.logging.info('* Testing failed.')
            return exit(1)

        disable_logging = True

        if self.args.logging in ('enable', 'true', '1', 'yes', 'on'):
            disable_logging = False

        tester.server_start(disable_logging=disable_logging)
        tested = tester.run()
        tester.server_stop()

        if not tested:
            self.logging.info('* Testing failed.')
        else:
            self.logging.info('* Testing succeed.')

        if not tested:
            return exit(1)

    def _append_sys_path(self, init_path=None):
        dirpath = init_path or self._pathified()
        modules = []

        while True:
            dirpath, module = self.helper.io.path.split(dirpath)
            modules.append(module)

            if not self.helper.io.path.is_file(self.helper.io.path.join(dirpath, '__init__.py')):
                break

        app_module = '.'.join(modules[::-1])

        self.helper.io.path.sys.insert(0, dirpath)
        self.helper.io.path.sys.insert(1, init_path)

        return app_module

    def _pathified(self):
        path = self.helper.io.path.join(self.cwd, self.args.path) if self.args.path else self.cwd
        path = path[:-1] if path.endswith('/') else path

        if path.endswith('__init__.py'):
            path = self.helper.io.path.dirname(path)

        return path

cli = CliHandler()


def main(as_module=False):
    cli.main()


if __name__ == '__main__':
    main(as_module=True)



# -*- coding: utf-8 -*-


import os
import requests
import subprocess
import tempfile

from dp_tornado.engine.cache import dpInValueModelConfig
from dp_tornado.engine.engine import Engine as dpEngine


class Compressor(dpEngine):
    def __init__(self, handler):
        self.handler = handler
        self.provider = {
            'minifier': 'minify'
        }

    def _commands(self, path, files, ext, tempname):
        files = [os.path.join(path, fn[1:] if fn.startswith('/') else fn) for fn in files]

        if ext == 'js':
            return [self.provider['minifier'], '--output', tempname] + files
        elif ext == 'css':
            return [self.provider['minifier'], '--output', tempname] + files

    def _compress(self, path, files, ext, tempname):
        tmpfiles = []
        offset = 0

        for file in files:
            offset += 1

            tmpname = '%s_%s' % (tempname, offset)
            tmp = open(tmpname, 'w')
            tmp.close()

            commands = self._commands(path, [file], ext, tmpname)
            subprocess.check_output(commands)

            tmpfiles.append(tmpname)

        with open(tempname, 'w') as outfile:
            for tmpfile in tmpfiles:
                with open(tmpfile) as infile:
                    for line in infile:
                        outfile.write(line)

                os.remove(tmpfile)

        return True

    def compress(self, path, files, ext, tempname, dp_static=False):
        try:
            path = path if not dp_static else self.handler.vars.static.path_dp

            if ext == 'js':
                commands = self._commands(path, files, ext, tempname)
                subprocess.check_output(commands)
            else:
                if not self._compress(path, files, ext, tempname):
                    return False

            return True

        except subprocess.CalledProcessError as e:
            self.logging.error('--------------------------------')
            self.logging.error('Static file minification error :')
            self.logging.error('Return code : %s' % e.returncode)
            self.logging.error('Command : %s' % e.cmd)
            self.logging.error('Output : %s' % e.output)
            self.logging.error('--------------------------------')
            for ee in files:
                self.logging.error(ee)
            self.logging.error('--------------------------------')

        except Exception as e:
            self.logging.error('--------------------------------')
            self.logging.error('Static file minification error :')
            self.logging.error(e)
            self.logging.error('--------------------------------')
            for ee in files:
                self.logging.error(ee)
            self.logging.error('--------------------------------')

        return False

    def _read(self, location):
        f = open(location, 'r')
        r = f.readlines()
        f.close()

        return ''.join(r)

    @staticmethod
    def clear(combined_dir):
        if not os.path.isdir(combined_dir):
            os.makedirs(combined_dir)

        for f in os.listdir(combined_dir):
            path = os.path.join(combined_dir, f)

            if not os.path.isdir(path):
                os.remove(path)


class StaticURL(dpEngine):
    def __init__(self, handler):
        self.handler = handler

        if not self.vars.server.start_at:
            self.vars.server.start_at = self.helper.datetime.timestamp.now()

    def render(self, *statics, **options):
        if not self.handler.vars.compressor:
            self.handler.vars.compressor = Compressor(self.handler)
            self.handler.vars.static.minify = self.ini.static.get('minify')
            self.handler.vars.static.path = self.ini.static.get('path_sys')
            self.handler.vars.static.prefix = self.ini.static.get('prefix')
            self.handler.vars.static.combined_url = self.ini.static.get('combined_url')
            self.handler.vars.static.combined_path = self.ini.static.get('combined_path')
            self.handler.vars.static.combined_prefix = self.ini.static.get('combined_prefix')
            self.handler.vars.static.cache_config = dpInValueModelConfig('sqlite', 'static_url', pure=True)

            path_dp = self.handler.helper.io.path.dirname(__file__)  # engine/plugin
            path_dp = self.handler.helper.io.path.dirname(path_dp)  # engine
            path_dp = self.handler.helper.io.path.join(path_dp, 'static')  # engine/static

            self.handler.vars.static.path_dp = path_dp

            self.handler.vars.static.aws_id = self.handler.ini.static.aws_id
            self.handler.vars.static.aws_secret = self.handler.ini.static.aws_secret
            self.handler.vars.static.aws_bucket = self.handler.ini.static.aws_bucket
            self.handler.vars.static.aws_region = self.handler.ini.static.aws_region
            self.handler.vars.static.aws_endpoint = self.handler.ini.static.aws_endpoint

            self.handler.vars.static.aws_configured = True if (self.handler.vars.static.aws_id and
                                                               self.handler.vars.static.aws_secret and
                                                               self.handler.vars.static.aws_bucket and
                                                               self.handler.vars.static.aws_region and
                                                               self.handler.vars.static.aws_endpoint) else False

        dp_static = True if 'dp' in options and options['dp'] else False
        dp_init = True if dp_static and 'init' in options and options['init'] else False

        if dp_init:
            if self.handler.request.headers and 'X-Proxy-Prefix' in self.handler.request.headers:
                dp_init = ["<script>var dp_prefix='%s';</script>" % self.handler.request.headers['X-Proxy-Prefix']]
            else:
                dp_init = []
        else:
            dp_init = []

        explicit = None
        explicit = 'debug' if 'debug' in options and options['debug'] else explicit
        explicit = 'skip' if 'skip' in options and options['skip'] else explicit
        explicit = 'local' if 'local' in options and options['local'] else explicit
        explicit = 's3' if 's3' in options and options['s3'] else explicit
        explicit_proxy = True if options and 'proxy' in options and options['proxy'] else False

        if explicit_proxy:
            explicit = None

        if explicit == 's3' and not self.handler.vars.static.aws_configured:
            explicit = 'local'
            self.handler.logging.warning('Required AWS configuration. explicit option is ignored.')

        # for Debugging Mode, skip minify css or js.
        if explicit == 'debug' or (not explicit and self.ini.server.debug):
            mtime = self.handler.vars.compressor.helper.datetime.timestamp.now(ms=True)
            return '\n'.join(dp_init + [self._template('%s?%s' % (t, mtime), dp_static=dp_static) for t in statics])

        elif explicit == 'skip' or (not explicit and (not self.handler.vars.static.minify or explicit_proxy)):
            return '\n'.join(dp_init + [self._template(
                '%s?%s' % (t, self.vars.server.start_at), dp_static=dp_static) for t in statics])

        cache_key = 'key_%s_%s' % (explicit, self.handler.helper.security.crypto.hash.sha224('/'.join(sorted(statics))))
        prepared = (self.handler.vars.static.prepared.__getattr__(cache_key) or
                    self.handler.cache.get(cache_key, dsn_or_conn=self.handler.vars.static.cache_config))

        if prepared:
            return prepared

        if '_separate' not in options or not options['_separate']:
            extensions = {}

            for static in statics:
                ext = static.split('?')[0].split('.')[-1].lower()

                if ext not in extensions:
                    extensions[ext] = []

                extensions[ext].append(static)

            if not extensions:
                return ''

            if len(extensions) > 1:
                options['_separate'] = False
                return '\n'.join(dp_init + [self.render(*static, **options) for static in list(extensions.values())])
            else:
                statics = list(extensions.values())[0]

        template = self._template(statics[0], replaced=False, dp_static=dp_static)
        ext = statics[0].split('.')[-1].lower()

        if not template:
            return ''

        acquire_key = 'job/%s' % cache_key
        acquire_identifier = self.handler.helper.misc.uuid.v1()

        if not self.handler.cache.get(acquire_key, dsn_or_conn=self.handler.vars.static.cache_config):
            self.handler.cache.set(
                acquire_key,
                acquire_identifier,
                dsn_or_conn=self.handler.vars.static.cache_config,
                expire_in=60*3)
        else:
            options['proxy'] = True
            return self.render(*statics, **options)

        filename = '%s.%s' % (self.handler.helper.security.crypto.hash.sha224(self.handler.helper.misc.uuid.v1()), ext)
        use_systmp = False

        if use_systmp:
            output = tempfile.NamedTemporaryFile(mode='w', prefix=filename, suffix='.%s' % ext, delete=False)
            tempname = output.name
            output.close()
        else:
            output = open(os.path.join(self.handler.vars.static.combined_path, filename), 'w')
            tempname = output.name
            output.close()

        compressed = self.handler.vars.compressor.compress(
            self.handler.vars.static.path, statics, ext, tempname, dp_static=dp_static)

        # If failed compression.
        if not compressed:
            options['proxy'] = True
            return self.render(*statics, **options)

        # AWS not configured, compressed files are stored local storage.
        if explicit == 'local' or not self.handler.vars.static.aws_configured:
            prepared = template % {
                'url': self.handler.vars.compressor.helper.web.url.join(
                    self.handler.vars.static.combined_prefix, filename)}

            test = os.path.exists(tempname)

        # AWS configured, upload compressed files to S3.
        else:
            content_length = os.path.getsize(tempname)
            filename = '%s/%s' % (self.handler.vars.static.combined_url, filename)

            self.handler.helper.web.aws.s3.upload(
                dest=filename,
                src=tempname,
                access_key_id=self.handler.vars.static.aws_id,
                secret_access_key=self.handler.vars.static.aws_secret,
                bucket_name=self.handler.vars.static.aws_bucket,
                region_name=self.handler.vars.static.aws_region,
                ExtraArgs={'ContentType': 'text/%s' % ('javascript' if ext == 'js' else ext)})

            os.remove(tempname)

            prepared_url = self.handler.vars.compressor.helper.web.url.join(
                self.handler.vars.static.aws_endpoint, filename)
            prepared = template % {'url': prepared_url}

            test = requests.get(prepared_url)
            test = False if test.status_code != 200 or (content_length and not len(test.text.strip())) else True

        if self.handler.cache.get(acquire_key, dsn_or_conn=self.handler.vars.static.cache_config) == acquire_identifier:
            if not test:
                options['proxy'] = True
                return self.render(*statics, **options)

            self.handler.vars.static.prepared.__setattr__(cache_key, prepared)
            self.handler.cache.set(cache_key, prepared, dsn_or_conn=self.handler.vars.static.cache_config)

        return prepared

    def _template(self, path, replaced=True, dp_static=False):
        if replaced:
            prefix = self.handler.vars.static.prefix if not dp_static else '/dp/'
            prefix = prefix if prefix.endswith('/') else '%s/' % prefix

            path = path[1:] if path.startswith('/') else path
            url = self.handler.vars.compressor.helper.web.url.join(prefix, path)

            return self._template(path, replaced=False) % {'url': url}

        extension = path.split('?')[0].split('.')[-1].lower()

        if extension == 'css':
            return '<link rel="stylesheet" type="text/css" href="%(url)s" />'
        elif extension == 'js':
            return '<script type="text/javascript" src="%(url)s"></script>'
        else:
            raise NotImplementedError



# -*- coding: utf-8 -*-


import importlib
import tornado.gen
import tornado.concurrent

from dp_tornado.engine.handler import Handler


class DpIdentifierHandler(Handler):
    def get(self, path=None):
        if self.remote_ip != '127.0.0.1':
            return self.finish_with_error(404)

        self.finish(self.ini.server.identifier or 'none')



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import hashlib


class HashHelper(dpHelper):
    def md5(self, plain):
        if self.helper.misc.type.check.numeric(plain):
            plain = str(plain)

        if not self.helper.misc.type.check.string(plain):
            raise Exception('The specified value is not plain text.')

        return hashlib.md5(str(plain).encode()).hexdigest()

    def sha224(self, plain):
        if self.helper.misc.type.check.numeric(plain):
            plain = str(plain)

        if not self.helper.misc.type.check.string(plain):
            raise Exception('The specified value is not plain text.')

        return hashlib.sha224(str(plain).encode()).hexdigest()



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper
from dp_tornado.engine.handler import Handler as dpHandler


class SessionHelper(dpHelper):
    def key(self, identifier):
        return 'session_%s' % (identifier.replace('-', '_'))

    def is_authorized(self, controller, identifier):
        return True if self.authorized(controller=controller, identifier=identifier) else False

    def authorize(self, controller, identifier, payload, after=None, before=None):
        if not isinstance(payload, dict):
            return False

        if before and before(payload) is False:
            return False

        controller.session(
            name=self.key(identifier),
            value=self.helper.serialization.json.stringify(payload),
            expire_in=self.ini.session.expire_in)

        if after and self.is_authorized(identifier=identifier, controller=controller):
            after(payload)

        return True

    def unauthorize(self, controller, identifier):
        controller.session(self.key(identifier), '', expire_in=0)
        return True

    def authorized(self, controller, identifier):
        controller = controller if isinstance(controller, dpHandler) else controller.parent

        # Return temporary session if exist
        if hasattr(controller, self.key(identifier)):
            return getattr(controller, self.key(identifier))

        payload = controller.session(name=self.key(identifier), expire_in=self.ini.session.expire_in)

        if not payload:
            return False

        payload = self.helper.serialization.json.parse(payload)

        # Save temporary session for current session
        if payload:
            setattr(controller, self.key(identifier), payload)

        return payload



# -*- coding: utf-8 -*-


import zipfile

from dp_tornado.engine.helper import Helper as dpHelper


class ZipHelper(dpHelper):
    def archive(self, destfile, srcfiles, mode='w', compression=zipfile.ZIP_STORED, allowZip64=False):
        archive = zipfile.ZipFile(file=destfile, mode=mode, compression=compression, allowZip64=allowZip64)

        if not isinstance(srcfiles, (list, tuple)):
            srcfiles = (srcfiles, )

        for srcfile in srcfiles:
            arcname = None
            compress_type = None

            if isinstance(srcfile, (tuple, list)):
                filename = srcfile[0]
                arcname = srcfile[1] if len(srcfile) > 1 else arcname
                compress_type = srcfile[2] if len(srcfile) > 2 else compress_type

            else:
                filename = srcfile

            self._archive_append(archive, filename, arcname, compress_type)

        archive.close()

        return True

    def _archive_append(self, archive, path, arcname, compress_type):
        if self.helper.io.path.is_file(path):
            archive.write(filename=path, arcname=arcname, compress_type=compress_type)
        elif self.helper.io.path.is_dir(path):
            for e in self.helper.io.path.browse(path):
                self._archive_append(archive=archive, path=e, arcname=arcname, compress_type=compress_type)

    def unarchive(self, srcfile, destpath, mode='r', compression=zipfile.ZIP_STORED, allowZip64=False):
        with zipfile.ZipFile(file=srcfile, mode=mode, compression=compression, allowZip64=allowZip64) as archive:
            archive.extractall(destpath)

        return True



# -*- coding: utf-8 -*-

"""

  dp for Tornado

  MVC Web Application Framework with Tornado
  http://github.com/why2pac/dp-tornado

  Copyright (c) 2015, why2pac <youngyongpark@gmail.com>

"""


import tornado.web
import tornado.ioloop
import tornado.httpserver

import time
import os
import multiprocessing
import importlib

from dp_tornado.engine.engine import EngineSingleton as dpEngineSingleton
from dp_tornado.engine.bootstrap import Bootstrap as EngineBootstrap
from dp_tornado.engine.scheduler import Scheduler
from dp_tornado.engine.plugin.static import Compressor
from dp_tornado.engine.plugin.static import StaticURL
from dp_tornado.engine.plugin.pagination import Pagination
from dp_tornado.engine.plugin import ui_methods
from dp_tornado.version import __version_info__


engine = dpEngineSingleton()


class RestfulApplication(tornado.web.Application):
    def __init__(self, handlers, kwargs):
        self.startup_at = int(round(time.time() * 1000))

        super(RestfulApplication, self).__init__(handlers, **kwargs)


class Bootstrap(object):
    def run(self, **kwargs):
        as_cli = kwargs['as_cli'] if 'as_cli' in kwargs and kwargs['as_cli'] else False
        dryrun = True if as_cli and as_cli.args.dryrun == 'yes' else False

        custom_scheduler = kwargs['scheduler'] if 'scheduler' in kwargs else None
        custom_service = kwargs['service'] if 'service' in kwargs else None
        custom_config_file = kwargs['config_file'] if 'config_file' in kwargs else 'config.ini'

        application_path = kwargs['application_path'] if 'application_path' in kwargs else None
        engine_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'engine')
        engine_static_path = os.path.join(engine_path, 'static')

        os.environ['DP_APPLICATION_PATH'] = application_path
        os.environ['DP_APPLICATION_INI'] = custom_config_file

        settings = EngineBootstrap.init_ini(
            application_path=application_path, ini_file=custom_config_file, as_cli=as_cli)

        engine.logger.sys_log('---------------------------------')
        engine.logger.sys_log('dp for Python            v%s' % '.'.join([str(e) for e in __version_info__]))
        engine.logger.sys_log('---------------------------------')

        services_raw = [
            (r"/dp/scheduler/(.*)", 'dp_tornado.engine.scheduler_handler.SchedulerHandler'),
            (r"/dp/identifier", 'dp_tornado.engine.dp_identifier_handler.DpIdentifierHandler'),
            (r"/dp/(.*)", 'dp_tornado.engine.static_handler.StaticHandler', {'path': engine_static_path}),
            (r"/", None),
            (r"/(.*)", None),
        ]

        if custom_service:
            services_raw = custom_service + services_raw

        services = []
        default_handler = None

        for service in services_raw:
            if len(service) < 2:
                raise Exception('The specified service is invalid.')

            if service[1] is not None:
                s = str.split(service[1], '.')
                class_name = s.pop()
                module_path = '.'.join(s)

                handler_module = importlib.import_module(module_path)
                handler = getattr(handler_module, class_name)

            else:
                if default_handler is None:
                    handler_module = importlib.import_module('dp_tornado.engine.default_handler')
                    default_handler = getattr(handler_module, 'DefaultHandler')

                module_path = 'controller'
                handler = default_handler

            services.append((service[0], handler, dict(prefix=module_path) if len(service) < 3 else service[2]))

        # Clear combined files
        Compressor.clear(settings['combined_static_path'])
        num_processed = engine.ini.server.num_processes if engine.ini.server.num_processes \
            else multiprocessing.cpu_count()

        deploy_mode = 'Production' if not engine.ini.server.debug else 'Debugging'
        identify_mode = engine.ini.app.mode

        engine.logger.sys_log('Server Mode : %s (%s)' % (deploy_mode, identify_mode))
        engine.logger.sys_log('Server time : %s' % time.strftime('%Y.%m.%d %H:%M:%S'))
        engine.logger.sys_log('Server Port : %s' % engine.ini.server.port)
        engine.logger.sys_log('Processors  : %s' % num_processed)
        engine.logger.sys_log('CPU Count   : %d' % multiprocessing.cpu_count())
        engine.logger.sys_log('---------------------------------')

        application = RestfulApplication(services, settings)
        service = tornado.httpserver.HTTPServer(
            application,
            xheaders=True,
            max_body_size=engine.ini.server.max_body_size)

        try:
            service.bind(engine.ini.server.port, '')
        except Exception as e:
            engine.logger.error('Failed to service binding. (port %s)' % engine.ini.server.port)
            engine.logger.error(e)
            return False

        if custom_scheduler:
            scheduler = Scheduler(custom_scheduler)
            scheduler.start()

        else:
            scheduler = None

        service.start(engine.ini.server.num_processes)

        import random
        application.identifier = random.randint(100000, 999999)

        engine.logger.start_handler()

        try:
            instance = tornado.ioloop.IOLoop.instance()
            instance.__setattr__('startup_at', getattr(application, 'startup_at'))

            if scheduler:
                scheduler.prepare()

            if not dryrun:
                instance.start()

        except KeyboardInterrupt:
            pass

        if scheduler:
            scheduler.interrupt()

        engine.logger.interrupt()



# -*- coding: utf-8 -*-


from dp_tornado.engine.config import Config


class DbConfig(Config):
    def index(self):
        self.conf.databases = {
            'service': {
                'driver': 'sqlite'
            }
        }



# -*- coding: utf-8 -*-
"""Testing module provides unittest by pre-defined testing syntax.
You can define testing syntax as method `docstring`.

Run Test:

.. testcode::

    $ dp4p test --path=./example

Available syntax:

* **expect(criteria)**
* **expect(priority, ruels)**
* **!expect(criteria)**
* **!expect(priority, ruels)**

.. testcode::

    ruels:
        controller:
            code, text, json, args(arguments separated by /), params(query string, form values)
        model, helper:
            int, long, bool, str, json, args (arguments by list), kwargs (arguments by dict)

.. testcode::

    class FooController(dpController):
        def get(self):
            \"""
                .. test::
                    expect(
                        1,
                        code=200,
                        text='foo==bar',
                        params={'foo': 'bar'})
            \"""
            foo = self.get_argument('foo')

            if foo == 'bar':
                return self.finish('foo==bar')

            self.finish('done')

        def post(self):
            \"""
                .. test::
                    expect(code=400, params={'foo': 'bar'})
            \"""
            foo = self.get_argument('foo')

            if foo == 'bar':
                return self.finish_with_error(400)

            self.finish('done')

    class FooController(dpController):
        def get(self):
            \"""
                Test with params and expect json value.

                .. test::
                    expect(json="{'foo':'bar'}", params={'foo': 'bar'})
            \"""
            foo = self.get_argument('foo')
            return self.finish({'foo': bar)

    class FooController(dpController):
        def get(self):
            \"""
                Test with params and expect json value.

                .. test::
                    expect(json={'foo':'bar'}, params={'foo': 'bar'})
            \"""
            foo = self.get_argument('foo')
            return self.finish({'foo': bar)

    class FooModel(dpModel):
        def foobar(self, foo):
            \"""
                Test with kwargs arguments.

                .. test::
                    expect(int=100, kwargs={'foo': 'bar'})
            \"""
            if foo == 'bar':
                return int(100)

            return 0

    class FooModel(dpModel):
        def foobar(self, foo):
            \"""
                Test with kwargs arguments.

                .. test::
                    expect(long=100, kwargs={'foo': 'bar'})
                    expect(long=0, kwargs={'foo': 'foo'})
                    !expect(long=0, kwargs={'foo': 'bar'})
            \"""
            if foo == 'bar':
                if self.e.helper.misc.system.py_version <= 2:
                    return long(100)
                else:
                    return int(100)


            if self.e.helper.misc.system.py_version <= 2:
                return long(0)
            else:
                return int(0)

    class FooHelper(dpHelper):
        def foobar(self, foo, bar):
            \"""
                Test with args arguments.

                .. test::
                    expect(bool=True, args={'foo': 'bar'})
                    !expect(bool=False, args={'foo': 'bar'})
            \"""
            foo = self.get_argument('foo')

            if foo == 'bar':
                return True

            return False
"""


import os
import sys
import time
import subprocess

from .engine import Engine as dpEngine


class Testing(dpEngine):
    def __init__(self, app_module, path):
        self.app_module = app_module
        self.app_identifier = 'dp-tornado-testing-9x48923'
        self.app_port = 48923

        self.path = path

        self.tests = {}
        self.modules = {}

        for e in ('controller', 'model', 'helper'):
            self.tests[e] = []

    @property
    def dev_null(self):
        return open(os.devnull, 'w')

    def server_start(self, disable_logging=True):
        self.server_stop()

        kwargs = {}

        if disable_logging:
            kwargs = {
                'stdout': self.dev_null,
                'stderr': subprocess.STDOUT
            }

        subprocess.Popen(
            ['dp4p', 'run', '--path', self.path, '--port', str(self.app_port), '--identifier', self.app_identifier],
            **kwargs)

    def server_stop(self):
        pids = subprocess.Popen(['pgrep', '-f', self.app_identifier], stdout=subprocess.PIPE)
        pids = pids.stdout.readlines()
        pids = [(e.decode('utf8') if sys.version_info[0] >= 3 else e).replace('\n', '') for e in (pids if pids else [])]

        for pid in pids:
            subprocess.Popen(
                ['kill', '-9', pid],
                stdout=self.dev_null,
                stderr=subprocess.STDOUT)

    def traverse(self):
        for module in ('controller', 'model', 'helper'):
            if not self._traverse(module, self.helper.io.path.join(self.path, module)):
                return False

            self.tests[module].sort(key=lambda e: e[6])

        return True

    def _traverse(self, module, path):
        for e in self.helper.io.path.browse(path, fullpath=False, recursive=True, conditions={'ext': 'py'}):
            path, cls = self.import_class(module, e)

            if cls is False:
                return False

            priority = 1000000

            for m in dir(cls):
                attr = getattr(cls, m)
                docstring = attr.__doc__

                if docstring and docstring.find('.. test::') != -1:
                    # noinspection PyUnusedLocal
                    def expect(alt, prio, *args, **kwargs):
                        # Priority
                        if args and len(args) == 1:
                            prio = args[0]

                        self.tests[module].append((cls, m, alt, path, kwargs, module, prio))

                    docstring = docstring[docstring.find('.. test::')+len('.. test::'):]
                    docstring = '\n'.join([e for e in docstring.split('\n') if e.strip()]).strip()

                    while True:
                        if docstring.find('expect(') != 0 and docstring.find('!expect(') != 0:
                            break

                        next_stmt = docstring[2:].find('expect(')

                        stmt = docstring[:next_stmt].strip() if next_stmt != -1 else docstring
                        docstring = docstring[len(stmt):].strip()

                        priority += 1

                        stmt = stmt.replace('expect(', 'expect(True, %s, ' % priority)
                        stmt = stmt.replace('!expect(True,', 'expect(False,')

                        # noinspection PyBroadException
                        try:
                            eval(stmt)
                        except Exception:
                            self.logging.info('* Test case parsing error, %s' % stmt)
                            return False

        return True

    def import_class(self, mod, path):
        path = path.split('.')[0].split('/')

        if path[-1] == '__init__':
            path.pop()

        if mod in ('model', 'helper') and not path:
            return path, None

        cls_name = ''.join([e.capitalize() for e in (path[-1] if path else 'starter').split('_')] + [mod.capitalize()])
        module = '.'.join([self.app_module, mod] + path)

        # noinspection PyBroadException
        try:
            __import__(module)
            app = sys.modules[module] if module in sys.modules else None
            cls = getattr(app, cls_name, None)

            if not cls:
                self.logging.info('* Class load error, %s' % module)
                return path, False

            return path, cls

        except:
            self.logging.info('* File import error, %s' % module)

            return path, False

    def run(self):
        self.logging.set_level('requests', self.logging.level.CRITICAL)

        server_executed = False

        for i in range(2*10):
            time.sleep(0.2)

            code, res = self.helper.web.http.get.text('http://127.0.0.1:%s/dp/identifier' % self.app_port)

            if res == self.app_identifier:
                server_executed = True
                break

            if i >= 5:
                self.logging.info('* Waiting server ..')

            time.sleep(0.3)

        if not server_executed:
            self.logging.info('* Server execution failed.')
            return exit(1)

        self.logging.info('*')

        session = True

        for e in self.tests['controller']:
            session, asserted = self._test_request(e, session)

            if not asserted:
                return False

        self.logging.info('*')

        for e in self.tests['model']:
            if not self._test_value(e):
                return False

        self.logging.info('*')

        for e in self.tests['helper']:
            if not self._test_value(e):
                return False

        self.logging.info('*')

        return True

    def _test_request(self, p, session):
        url_path = '/'.join(p[3])
        path = '%s.%s.%s' % (p[5], '.'.join(p[3]), p[1])
        method = p[1]
        req = {}
        res_type = None

        if 'args' in p[4] and p[4]['args']:
            url_path = '%s/%s' % (url_path, '/'.join(str(e) for e in p[4]['args']))

        url = 'http://127.0.0.1:%s/%s' % (self.app_port, url_path)

        if 'params' in p[4] and p[4]['params']:
            req['data'] = p[4]['params']

        if 'code' in p[4]:
            res_type = 'raw'
        if 'text' in p[4]:
            res_type = 'text'
        elif 'json' in p[4]:
            res_type = 'json'

        if method not in ('get', 'post', 'delete', 'patch', 'put', 'head') or not res_type:
            self.logging.info('* Method test, %s -> (%s) [INVALID]' % (path, req or '-'))
            return session, False

        session, code, res = self.helper.web.http.request(
            req_type=method, res_type=res_type, url=url, session=session or True, **req)

        asserted_code = None
        asserted_text = None
        asserted_json = None

        # Assertion, code
        if 'code' in p[4]:
            asserted_code = True if p[4]['code'] == code else False

        # Assertion, text
        if 'text' in p[4]:
            if res_type != 'text':
                res = str(res)

            asserted_text = True if p[4]['text'] == res else False

        # Assertion, json
        if 'json' in p[4]:
            res_a = res if self.helper.misc.type.check.string(res) else self.helper.serialization.json.stringify(res)
            res_a = self.helper.serialization.json.parse(res_a)

            res_b = self.helper.serialization.json.stringify(p[4]['json'])
            res_b = self.helper.serialization.json.parse(res_b)

            asserted_json = True if res_a == res_b else False

        asserted = asserted_code is False or asserted_text is False or asserted_json is False

        if p[2]:
            asserted = not asserted

        if not asserted:
            desc = []

            if asserted_code is not None:
                desc.append('[CODE %s / %s]' % (p[4]['code'], code))
            if asserted_text is not None:
                desc.append('[TEXT "%s" / "%s"]' % (p[4]['text'], res))
            if asserted_json is not None:
                # noinspection PyUnboundLocalVariable
                res_a = self.helper.serialization.json.stringify(res_a)
                # noinspection PyUnboundLocalVariable
                res_b = self.helper.serialization.json.stringify(res_b)

                res_a = '%s..' % res_a[0:7] if len(res_a) > 7 else res_a
                res_b = '%s..' % res_b[0:7] if len(res_b) > 7 else res_b

                desc.append('[JSON %s / %s]' % (res_a, res_b))

            desc = ' & '.join(desc)

            self.logging.info(
                '* Request test, [%s] %s => %s %s [FAIL]' % (method.upper(), url_path, '' if p[2] else '!', desc))

            return session, False

        desc = []

        if asserted_code is not None:
            desc.append('CODE')
        if asserted_text is not None:
            desc.append('TEXT')
        if asserted_json is not None:
            desc.append('JSON')

        self.logging.info(
            '* Request test, [%s] %s -> %s [OK]' % (method.upper(), url_path, ' & '.join(desc)))

        return session, True

    def _test_request_assertion(self, payload, result):
        pass

    def _test_value(self, p):
        cls = self._class(p[0])
        method = getattr(cls, p[1])
        path = '%s.%s.%s' % (p[5], '.'.join(p[3]), p[1])
        req = None

        # noinspection PyBroadException
        try:
            if 'kwargs' in p[4] and p[4]['kwargs']:
                req = p[4]['kwargs']
                got = method(**req)
            elif 'args' in p[4] and p[4]['args']:
                req = p[4]['args']
                got = method(*req)
            else:
                got = method()
        except:
            self.logging.info('* Method execution error, %s.%s' % (p[0], p[1]))
            return False

        res, exp = self._test_value_assertion(p, got)

        if not res:
            self.logging.info(
                '* Method test, %s -> (%s) -> %s%s -> %s [FAIL]' % (path, req, '' if p[2] else '! ', got, exp))
            return False

        got = str(got)
        req = str(req)

        got = '%s..' % got[0:7] if len(got) > 7 else got
        req = '%s..' % req[0:7] if len(req) > 7 else req

        self.logging.info(
            '* Method test, %s -> (%s) -> %s%s [OK]' % (path, req, '' if p[2] else '! ', got))

        return True

    def _test_value_assertion(self, payload, result):
        expected = {}

        for k in ('int', 'long', 'bool', 'str', 'json'):
            if k in payload:
                expected[k] = payload[k]

        for k, v in expected.items():
            vo = v

            if k == 'json':
                vo = self.helper.serialization.json.stringify(vo)
                vo = self.helper.serialization.json.parse(vo)

                result = self.helper.serialization.json.stringify(result)
                result = self.helper.serialization.json.parse(result)

            if payload[2]:
                if vo != result:
                    return False, v
            else:
                if vo == result:
                    return False, v

        return True, None

    def _class(self, cls):
        if str(cls) in self.modules:
            return self.modules[str(cls)]

        self.modules[str(cls)] = cls()
        return self.modules[str(cls)]


# -*- coding: utf-8 -*-
"""
MVC Web Application Framework with Tornado, Python 2 and 3

To install the package run:

    pip install dp-tornado


Run
-----

    $ pip install virtualenv
    $ virtualenv ./venv
    $ . ./venv/bin/activate
    $ pip install dp-tornado
    $ dp4p init --path app

"""


import sys
import logging


try:
    from setuptools import setup

except ImportError:
    from distutils.core import setup

try:
    import distutils.core as core
except ImportError:
    core = None

try:
    from setuptools.command.install import install

except ImportError:
    from distutils.command.install import install


with open('dp_tornado/version.py', 'r') as fp_v:
    for c in fp_v.read().split('\n'):
        exec(c)


dp_project_name = 'dp-tornado'
dp_version = __version__
dp_github_url = 'http://github.com/why2pac/dp-tornado'
dp_license = 'MIT'

dp_author = 'Parker'
dp_author_email = 'oss@dp.farm'

dp_maintainer = dp_author
dp_maintainer_email = dp_author_email

dp_description = 'MVC Web Application Framework with Tornado.'

dp_requires_CyMySQL = 'CyMySQL==0.8.9'
dp_requires_futures = 'futures==3.0.5'


class CustomInstallCommand(install):
    user_options = install.user_options + [
        ('dp-identifier=', None, 'Specify identifier for dp.'),
        ('dp-without-mysql', None, 'Specify this option if you do not want to install mysql dependency.')
    ]

    def initialize_options(self):
        install.initialize_options(self)

        self.dp_identifier = None
        self.dp_without_mysql = False

    def finalize_options(self):
        install.finalize_options(self)

        dist = getattr(core, '_setup_distribution', None) if core else None

        if not dist:
            logging.warning('Aditional requires cannot installed.')
        else:
            if self.dp_without_mysql:
                dist.install_requires.remove(dp_requires_CyMySQL)

    def run(self):
        install.run(self)
        install.do_egg_install(self)


install_requires = [
        'tornado==4.4.2',
        'redis==2.10.5',
        'requests==2.12.3',
        'croniter==0.3.13',
        'pytz==2016.10',
        'pycrypto==2.6.1',
        'boto3==1.4.2',
        'SQLAlchemy==1.1.4',
        'Pillow==3.4.2',
        'validate_email==1.3',
        'BeautifulSoup4==4.5.1',
        'lxml==3.6.4',
        'httpagentparser==1.7.8',
        'validators==0.11.1',
        dp_requires_CyMySQL,
        # , 'selenium'
    ]


if sys.version_info[0] <= 2:
    install_requires.append(dp_requires_futures)


setup(
    name=dp_project_name,
    version=dp_version,
    url=dp_github_url,
    license=dp_license,
    author=dp_author,
    author_email=dp_author_email,
    maintainer=dp_maintainer,
    maintainer_email=dp_maintainer_email,
    description=dp_description,
    long_description=__doc__,
    packages=['dp_tornado'],
    include_package_data=True,
    cmdclass={'install': CustomInstallCommand},
    install_requires=install_requires,
    keywords=['MVC', 'Web Application Framework'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4'
    ],
    entry_points={
        'console_scripts': [
            'dp4p = dp_tornado.cli:main'
        ]
    }
)



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import base64


class EncodingHelper(dpHelper):
    def base64_encode(self, plain, raw=False):
        if not raw and self.helper.misc.system.py_version >= 3 and self.helper.misc.type.check.string(plain):
            plain = bytes(plain, 'utf8')

        encoded = base64.standard_b64encode(plain)

        if self.helper.misc.system.py_version >= 3:
            encoded = encoded.decode('utf8')

        return encoded

    def base64_decode(self, encoded, raw=False):
        if not raw and self.helper.misc.system.py_version >= 3 and self.helper.misc.type.check.string(encoded):
            encoded = bytes(encoded, 'utf8')

        decoded = base64.standard_b64decode(encoded)

        if not raw and self.helper.misc.system.py_version >= 3:
            decoded = decoded.decode('utf8')

        return decoded



# -*- coding: utf-8 -*-


import time

from dp_tornado.engine.helper import Helper as dpHelper


class TableHelper(dpHelper):
    def get(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            wait_until_exists=False):
        dynamodb = self.helper.web.aws.dynamodb.resource(
            access_key_id=access_key_id, secret_access_key=secret_access_key, region_name=region_name)

        table = dynamodb.Table(table_name)

        if wait_until_exists:
            if not self.wait_until(
                    access_key_id=access_key_id,
                    secret_access_key=secret_access_key,
                    region_name=region_name,
                    table_name=table_name,
                    exists=True):
                return False

        return table

    def create(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            index_columns,
            read_capacity_units=1,
            write_capacity_units=1,
            wait_until_exists=True,
            **kwargs):
        try:
            dynamodb = self.helper.web.aws.dynamodb.client(
                access_key_id=access_key_id, secret_access_key=secret_access_key, region_name=region_name)

            def_at = self.helper.web.aws.dynamodb.table.indexing.sort

            key_schema = [{'AttributeName': e[0], 'KeyType': e[2] if len(e) >= 3 else def_at} for e in index_columns]
            attr_defs = [{'AttributeName': e[0], 'AttributeType': e[1]} for e in index_columns]

            dynamodb.create_table(
                TableName=table_name,
                KeySchema=key_schema,
                AttributeDefinitions=attr_defs,
                ProvisionedThroughput={
                    'ReadCapacityUnits': read_capacity_units,
                    'WriteCapacityUnits': write_capacity_units
                },
                **kwargs)

            if wait_until_exists:
                if not self.wait_until(
                        access_key_id=access_key_id,
                        secret_access_key=secret_access_key,
                        region_name=region_name,
                        table_name=table_name,
                        exists=True):
                    return False

            return True

        except Exception as e:
            if self.status(
                    access_key_id=access_key_id,
                    secret_access_key=secret_access_key,
                    region_name=region_name,
                    table_name=table_name) == 'ACTIVE':
                return True

            self.logging.exception(e)

            return False

    def wait_until(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            exists,
            repeat=25,
            repeated=0,
            before_status=None):
        try:
            status = self.status(
                access_key_id=access_key_id,
                secret_access_key=secret_access_key,
                region_name=region_name,
                table_name=table_name)

            if not status:
                if not exists and before_status in ('ACTIVE', 'DELETING'):
                    return True

                return False

            if exists and status == 'ACTIVE':
                return True

            if repeat > repeated:
                time.sleep(0.3)

                return self.wait_until(
                    access_key_id=access_key_id,
                    secret_access_key=secret_access_key,
                    region_name=region_name,
                    table_name=table_name,
                    exists=exists,
                    repeat=repeat,
                    repeated=repeated+1,
                    before_status=status)

            return True

        except Exception as e:
            self.logging.error(e)

            return False

    def status(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name):
        described = self.describe(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            region_name=region_name,
            table_name=table_name,
            wait_until_exists=False)

        if not described or 'Table' not in described or 'TableStatus' not in described['Table']:
            return False

        return described['Table']['TableStatus']

    def describe(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            wait_until_exists=True):
        try:
            if wait_until_exists:
                if not self.wait_until(
                        access_key_id=access_key_id,
                        secret_access_key=secret_access_key,
                        region_name=region_name,
                        table_name=table_name,
                        exists=True):
                    return False

            dynamodb = self.helper.web.aws.dynamodb.client(
                access_key_id=access_key_id, secret_access_key=secret_access_key, region_name=region_name)

            describe = dynamodb.describe_table(TableName=table_name)

            return describe

        except Exception as e:
            if wait_until_exists:
                self.logging.error(e)

            return False

    def remove(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            wait_until_not_exists=True):
        try:
            dynamodb = self.helper.web.aws.dynamodb.client(
                access_key_id=access_key_id, secret_access_key=secret_access_key, region_name=region_name)

            deleted = dynamodb.delete_table(TableName=table_name)

            if not deleted:
                return False

            if wait_until_not_exists:
                if not self.wait_until(
                        access_key_id=access_key_id,
                        secret_access_key=secret_access_key,
                        region_name=region_name,
                        table_name=table_name,
                        exists=False,
                        before_status='ACTIVE'):
                    return False

            return True

        except Exception as e:
            self.logging.exception(e)

            return False



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import uuid


class UuidHelper(dpHelper):
    @property
    def namespace_dns(self):
        return uuid.NAMESPACE_DNS

    @property
    def namespace_url(self):
        return uuid.NAMESPACE_URL

    @property
    def namespace_oid(self):
        return uuid.NAMESPACE_OID

    @property
    def namespace_x500(self):
        return uuid.NAMESPACE_X500

    def v1(self):  # make a UUID based on the host ID and current time
        return str(uuid.uuid1())

    def v3(self, namespace, name):
        return str(uuid.uuid3(namespace, name))

    def v4(self):
        return str(uuid.uuid4())

    def v5(self, namespace, name):
        return str(uuid.uuid5(namespace, name))



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class CastHelper(dpHelper):
    def int(self, a, raise_exception=False):
        try:
            if self.helper.misc.system.py_version <= 2:
                return int(a) if a else 0
            else:
                return int(a) if a else 0

        except ValueError as e:
            if raise_exception:
                raise e

            return False

    def long(self, a, raise_exception=False):
        try:
            if self.helper.misc.system.py_version <= 2:
                return long(a) if a else long(0)
            else:
                return int(a) if a else 0

        except ValueError as e:
            if raise_exception:
                raise e

            return False

    def float(self, a, raise_exception=False):
        try:
            return float(a) if a else 0

        except ValueError as e:
            if raise_exception:
                raise e

            return False
